import os
from pathlib import Path
from typing import Callable, Literal
from urllib.parse import parse_qsl
import warnings

import altair as alt
import polars as pl
import solara
import solara.lab
from hdxms_datasets import DataBase
from ipymolstar import PDBeMolstar
from ipymolstar.pdbemolstar import THEMES
from solara.alias import rv

from instagibbs.config import cfg
from instagibbs.plot.alt import alt_peptides, alt_residues
from instagibbs.plot.structure import pdbemolstar_colors, pdbemolstar_tooltips
from instagibbs.preprocess import load_hdx_dataset
from instagibbs.quantities import DDG, DRFU, RFU, PlotQty
from instagibbs.web.methods import (
    default_grid_layout,
    find_free_grid_position,
    new_card,
    td,
    th,
    tr,
)
from instagibbs.web.models import DataManager
from instagibbs.web.constants import BASE_DIR, IN_MEMORY_DATASETS, about_md

try:
    from dotenv import load_dotenv

    load_dotenv()
except ImportError:
    pass

DEV_MODE = os.environ.get("INSTAGIBBS_DEV", "0") == "1"

if DEV_MODE:
    warnings.warn("Running in DEV mode!")

dprint = print if DEV_MODE else lambda *args, **kwargs: None

# Make ipymolstar use white background by in light theme
THEMES["light"]["bg_color"] = "#FFFFFF"


def is_running_in_huggingface():
    return "SPACE_ID" in os.environ


# Set database directory based on environment
if DEV_MODE:
    database_dir = Path("C:/Users/jhsmi/repos/mine/HDXMS-database/incoming/staging/output")
# TODO remove in favour of config / CLI option
elif is_running_in_huggingface():
    database_dir = Path.cwd().parent / "HDXMS-database" / "datasets"
else:
    database_dir = cfg.database_dir
    database_dir.mkdir(exist_ok=True, parents=True)

DATABASE_DIR = solara.Reactive(database_dir)

# move to constants module
initial_types = ["peptides_chart", "residues_chart", "structure"]
title_mapping = {
    "peptides_chart": "Peptides Chart",
    "residues_chart": "Residues Chart",
    "structure": "Structure View",
    "peptides_table": "Peptides Table",
    "residues_table": "Residues Table",
}

view_values = [{"text": v, "value": k} for k, v in title_mapping.items()]
pin_cards = solara.Reactive(False)

views_type = Literal[
    "peptides_chart", "residues_chart", "structure", "peptides_table", "residues_table"
]


def apply_quantity_scaling(df: pl.DataFrame, qty: PlotQty) -> pl.DataFrame:
    cols = [(pl.col("value") * qty.scale_factor).alias(qty.label)]
    drop_cols = ["value"]
    if "value_sd" in df.columns:
        cols.append((pl.col("value_sd") * qty.scale_factor).alias(f"{qty.label} SD"))
        drop_cols.append("value_sd")
    return df.with_columns(cols).drop(drop_cols)


# %%
@solara.component  # type: ignore
def ViewCard(
    data_manager: DataManager,
    view_type: views_type | solara.Reactive[views_type] = "peptides_chart",
    width="100%",
    height="400px",
    on_delete: Callable[[], None] | None = None,
):
    local_view_type = solara.use_reactive(view_type)
    title = title_mapping.get(local_view_type.value, "View Card")
    dark_effective = solara.lab.use_dark_effective()
    sync_with_global = solara.use_reactive(True)

    # TODO element portal / check reuse
    AltairThemeSetter()

    local_dm = solara.use_memo(lambda: data_manager.clone(), dependencies=[])  # type: ignore
    dm = data_manager if sync_with_global.value else local_dm
    margin = 2
    with rv.Card(
        style_=f"height:{height}; width:{width};",
        class_=f"ma-{margin}",
        elevation=2,
    ):
        # TODO split style into where its needed?
        solara.Style(BASE_DIR / "style.css")
        with rv.CardTitle(
            style_="display: flex; justify-content: space-between; align-items: center;",
            class_="card card--flex",
        ):
            solara.Text(title)
            with solara.Row():
                # how to menu + tooltip?
                # solara.Tooltip("Change view type", children=[dashboard_btn])
                dashboard_btn = solara.IconButton("mdi-view-dashboard", on_click=lambda: None)
                with solara.lab.Menu(activator=dashboard_btn, close_on_content_click=False):
                    with solara.Div(class_="card-menu-content"):
                        solara.Select(
                            "View Type",
                            values=view_values,  # type: ignore
                            value=local_view_type,  # type: ignore
                            dense=False,
                        )
                is_peptide_view = local_view_type.value in ["peptides_table", "peptides_chart"]
                reduction_btn = solara.IconButton(
                    "mdi-alpha-r-circle",
                    on_click=lambda: None,
                    tooltip="Change residue reduction",
                    disabled=is_peptide_view,
                )
                with solara.lab.Menu(activator=reduction_btn, close_on_content_click=False):
                    with solara.Div(class_="card-menu-content"):
                        regression_option = (
                            "lasso_regression" if dm.is_delta else "ridge_regression"
                        )
                        solara.Select(  # type: ignore
                            "Residue Reduction",
                            values=["weighted_average", regression_option],
                            value=dm.residue_reduction.value,
                            on_value=dm.residue_reduction.set,  # type: ignore
                            dense=False,
                        )
                        solara.InputFloat(
                            label="Alpha",
                            value=dm.alpha,
                            disabled=dm.residue_reduction.value == "weighted_average",
                        )

                btn = solara.IconButton("mdi-tune-vertical", on_click=lambda: None)
                with solara.lab.Menu(activator=btn, close_on_content_click=False):
                    with solara.Div(class_="card-menu-content"):
                        solara.Select(
                            "Quantity",
                            values=dm.qty_values,
                            value=dm.qty.value.name,
                            on_value=dm.set_qty,
                            dense=False,
                        )
                        solara.Select("Base State", value=dm.base_state, values=dm.state_values)
                        enable_diff = dm.qty.value in [DDG, DRFU]
                        solara.Select(
                            "Diff State",
                            value=dm.diff_state,
                            values=dm.state_values,
                            disabled=not enable_diff,
                        )
                        solara.Select(
                            "Exposure",
                            value=dm.exposure,
                            values=dm.exposure_values,
                            disabled=dm.qty.value not in [RFU, DRFU],
                        )
                with solara.Tooltip("Toggle global controls sync"):
                    solara.IconButton(
                        "mdi-link-off" if not sync_with_global.value else "mdi-link",
                        on_click=lambda: sync_with_global.set(not sync_with_global.value),
                        color="accent" if not sync_with_global.value else None,
                    )

                if on_delete is not None:
                    # TODO fix tooltip
                    solara.IconButton("mdi-delete", on_click=on_delete, tooltip="Delete View")  # type: ignore

        rv.CardSubtitle(children=[dm.label])  # need more margin if subtitle is used

        margin = 80
        margin += 36  # account for subtitle
        parent_height = int(height.strip("px"))
        # this Div is for pdbemolstar element only, consider moving it there
        with solara.Div(
            style={
                "width": "calc(100% - 0px)",
                "height": f"calc(100% - {margin}px)",  # not sure if this height does anything right now
                "position": "relative",
                "padding": "0 16px",  # 16px on each side = 32px total
                "box-sizing": "border-box",  # Include padding in width calculation
            },
        ):
            if local_view_type.value == "peptides_table":
                df = apply_quantity_scaling(dm.get_peptide_df(), dm.qty.value)
                solara.DataFrame(df)
            if local_view_type.value == "residues_table":
                df = apply_quantity_scaling(dm.get_residue_df(), dm.qty.value)
                solara.DataFrame(df)
            elif local_view_type.value == "peptides_chart":
                chart = alt_peptides(
                    dm.get_peptide_df(), qty=dm.qty.value, height=parent_height - margin
                )
                chart = chart.configure(
                    padding=5,
                    autosize=alt.AutoSizeParams(type="fit", contains="content", resize=True),
                )
                # embed_options = {
                #     # "mode": "vega-lite",
                #     # "renderer": "svg",
                #     # "theme": "dark" if dark_effective else "default",
                #     "actions": False,
                # }
                # alt.JupyterChart.element(chart=chart.interactive(), embed_options=embed_options)  # type: ignore
                solara.FigureAltair(chart.interactive()).key(f"peptide_chart-{dark_effective}")

            elif local_view_type.value == "residues_chart":
                chart = alt_residues(
                    dm.get_residue_df(), qty=dm.qty.value, height=parent_height - margin, pad=0.1
                )
                chart = chart.configure(
                    padding=5,
                    autosize=alt.AutoSizeParams(type="fit", contains="content", resize=True),
                )
                # alt.JupyterChart.element(
                #     chart=chart
                # )  # , embed_options=embed_options)  # type: ignore
                solara.FigureAltair(chart).key(f"residue_chart-{dark_effective}")

            elif local_view_type.value == "structure":
                color_data = pdbemolstar_colors(
                    dm.get_residue_df(), qty=dm.qty.value, mapping=dm.structure_mapping
                )
                tooltips = pdbemolstar_tooltips(
                    dm.get_residue_df(), qty=dm.qty.value, mapping=dm.structure_mapping
                )
                PDBeMolstar.element(  # type: ignore
                    custom_data=dm.pdbe_molstar_custom_data,
                    hide_water=True,
                    color_data=color_data,
                    tooltips=tooltips,
                    height=f"{parent_height - margin}px",
                    # height=height,
                    theme="light" if not dark_effective else "dark",
                ).key(f"PDBEmolstar-{dark_effective}")


@solara.component  # type: ignore
def Sidebar(
    view_cards: solara.Reactive[list[dict]],
    grid_layout: solara.Reactive[list[dict]],
    global_dm: DataManager,
):
    new_view_type = solara.use_reactive(view_values[0]["value"])
    dprint("rerender in sidebar")

    # TODO remove
    def is_synced_reactive(reactive: str) -> bool:
        return True
        # return all(
        #     getattr(view_card["dm"], reactive).value == getattr(global_dm, reactive).value
        #     for view_card in view_cards.value
        # )

    with solara.Sidebar():
        with solara.Card("Controls"):
            solara.Select(
                "Quantity",
                values=global_dm.qty_values,
                value=global_dm.qty.value.name,
                on_value=global_dm.set_qty,
                dense=False,
                classes=["out-of-sync"] if not is_synced_reactive("qty") else [],
            )
            solara.Select(
                "Base State",
                value=global_dm.base_state,
                values=global_dm.state_values,
                classes=["out-of-sync"] if not is_synced_reactive("base_state") else [],
            )
            solara.Select(
                "Diff State",
                value=global_dm.diff_state,
                values=global_dm.state_values,
                disabled=not global_dm.is_delta,
                classes=["out-of-sync"] if not is_synced_reactive("diff_state") else [],
            )
            solara.Select(
                "Exposure",
                value=global_dm.exposure,
                values=global_dm.exposure_values,
                disabled=global_dm.qty.value not in [RFU, DRFU],
                classes=["out-of-sync"] if not is_synced_reactive("exposure") else [],
            )

            regression_option = "lasso_regression" if global_dm.is_delta else "ridge_regression"
            solara.Select(  # type: ignore
                "Residue Reduction",
                values=["weighted_average", regression_option],
                value=global_dm.residue_reduction.value,
                on_value=global_dm.residue_reduction.set,  # type: ignore
                dense=False,
                classes=["out-of-sync"] if not is_synced_reactive("residue_reduction") else [],
            )
            solara.InputFloat(
                label="Alpha",
                value=global_dm.alpha,
                disabled=global_dm.residue_reduction.value == "weighted_average",
                classes=["out-of-sync"] if not is_synced_reactive("alpha") else [],
            )
        with solara.Card("Add View"):

            def add_view():
                """Add a new view card with the specified view type."""
                new_card_entry = new_card(new_view_type.value)
                new_cards = view_cards.value[:] + [new_card_entry]

                new_layout = find_free_grid_position(grid_layout.value)
                if new_layout is None:
                    return  # No free position found, do not add a new view
                view_cards.set(new_cards)
                grid_layout.set(grid_layout.value[:] + [new_layout])

            with solara.Row():
                solara.Select(
                    "View Type",
                    value=new_view_type,  # type: ignore
                    values=view_values,  # type: ignore
                    dense=False,
                )
                solara.IconButton(
                    "mdi-plus",
                    on_click=add_view,
                )  # type: ignore


# %%
@solara.component  # type: ignore
def DataSetView(dataset_id: str):
    """Display a dataset view.

    Args:
        dataset: ID of the dataset to display
    """
    dprint("rerender in dataset_view")

    def load_dataset():
        local_db = DataBase(cfg.database_dir)
        # remote_db = RemoteDataBase(cfg.remote_url, database_dir=cfg.database_dir / "remote_cache")

        if dataset_id in IN_MEMORY_DATASETS:
            dataset = IN_MEMORY_DATASETS[dataset_id]
        elif dataset_id in local_db.datasets:
            dataset = local_db.load_dataset(dataset_id)
        # elif: TODO implement remote db (local cache)
        hdx_states = load_hdx_dataset(dataset)

        global_dm = DataManager(hdx_states=hdx_states)
        return global_dm

    global_dm = solara.use_memo(load_dataset, dependencies=[dataset_id])  # type: ignore

    def make_cards():
        view_cards = [new_card(view_type) for view_type in initial_types]
        initial_grid = default_grid_layout()

        return view_cards, initial_grid

    # use memo to run once
    initial_cards, initial_grid = solara.use_memo(make_cards, dependencies=[])  # type: ignore
    view_cards = solara.use_reactive(initial_cards)
    grid_layout = solara.use_reactive(initial_grid)

    Sidebar(view_cards, grid_layout, global_dm)

    def make_grid_card(layout, view_card):
        def on_delete(id=view_card["id"]):
            """Remove a view card by its ID."""
            card_ids = [card["id"] for card in view_cards.value]
            delete_idx = card_ids.index(id)
            to_remove = view_cards.value[delete_idx]

            if to_remove is None:
                return

            view_cards.value = [card for card in view_cards.value if card["id"] != id]

            new_layout, idx, moved = [], 0, False
            for i, layout in enumerate(grid_layout.value):
                if i == delete_idx:
                    moved = True
                    continue
                # adjust index, set items past the deleted card to 'moved' such they stay in place
                # TODO doesn't seem to work, resulting layout 'moved' is always False
                new_layout.append(layout | {"i": str(idx), "moved": moved})
                idx += 1

            grid_layout.value = new_layout

        height_pixels = layout["h"] * 40 - 10
        card = ViewCard(
            data_manager=global_dm,
            view_type=view_card["view_type"],
            width="100%",
            height=f"{height_pixels}px",
            on_delete=on_delete,
        )

        return card

    def make_grid_items():
        """Create grid items from view cards and their layouts."""
        return [
            make_grid_card(layout, view_card)
            for layout, view_card in zip(grid_layout.value, view_cards.value)
        ]

    grid_items = solara.use_memo(  # type: ignore
        make_grid_items, dependencies=[view_cards.value, grid_layout.value]
    )

    solara.GridDraggable(
        items=grid_items,
        grid_layout=grid_layout.value,
        resizable=True,
        draggable=not pin_cards.value,
        on_grid_layout=grid_layout.set,
        col_num=24,
    )


# used for appbar color
@solara.component  # type: ignore
def Layout(children):
    dark_effective = solara.lab.use_dark_effective()
    return solara.AppLayout(
        children=children,
        toolbar_dark=dark_effective,
        color="None",  # if dark_effective else "primary",
    )


def Home():
    db = DataBase(DATABASE_DIR.value)
    with solara.Card():
        solara.Markdown(
            "## Welcome to InstaGibbs!\nChoose an HDX-MS dataset from the table below to open in InstaGibbs."
        )

        with solara.v.SimpleTable():  # type: ignore
            with solara.v.Html(tag="thead"):  # type: ignore
                with tr():
                    th("Dataset")
                    th("Open in InstaGibbs")
                for d in db.datasets:
                    with tr():
                        td(d)
                        td(
                            solara.Link(
                                path_or_route=f"/?dataset_id={d}",
                                children=[solara.Button(icon_name="mdi-open-in-new", icon=True)],
                            )
                        )


# TODO use element portal to place this component at the root level
# and check if it already exists
@solara.component  # type: ignore
def AltairThemeSetter():
    dark_effective = solara.lab.use_dark_effective()

    def set_alt_theme():
        if dark_effective:
            alt.theme.enable("dark")
        else:
            alt.theme.enable("default")

    solara.use_memo(set_alt_theme, dependencies=[dark_effective])  # type: ignore
    return None


@solara.component  # type: ignore
def Main(dataset_id: str | None = None):
    """Main component for database mode."""
    settings_dialog = solara.use_reactive(False)
    about_dialog = solara.use_reactive(False)

    dark_effective = solara.lab.use_dark_effective()

    AltairThemeSetter()

    title = "InstaGibbs" + f" / {dataset_id}" if dataset_id is not None else ""

    solara.Title(title)
    with solara.AppBar():
        if dataset_id is not None:
            with solara.Tooltip("Pin view cards"):
                solara.Button(
                    icon_name="mdi-pin" if pin_cards.value else "mdi-pin-off",
                    icon=True,
                    on_click=lambda: pin_cards.set(not pin_cards.value),
                )

        with solara.Tooltip("Settings"):
            solara.Button(
                icon_name="mdi-settings",
                icon=True,
                on_click=lambda: settings_dialog.set(True),
            )

        with solara.Tooltip("About"):
            solara.Button(
                icon_name="mdi-information-outline",
                icon=True,
                on_click=lambda: about_dialog.set(True),
            )
        with solara.Tooltip("Toggle dark/light theme"):
            solara.lab.ThemeToggle()

        with solara.Tooltip("Home"):
            with solara.Link(path_or_route="/"):
                solara.Button(icon_name="mdi-home", icon=True)  # , classes=["mx-2"])

    if settings_dialog.value:
        with rv.Dialog(
            v_model=settings_dialog.value,
            on_v_model=settings_dialog.set,
            max_width=500,
        ):
            with solara.Card("Settings"):
                solara.Markdown("Settings is work in progress. ")

    if about_dialog.value:
        with rv.Dialog(
            v_model=about_dialog.value,
            on_v_model=about_dialog.set,
            max_width=500,
        ):
            with solara.Card("About instagibbs"):
                solara.Markdown(about_md)

    # TODO DATABASE_DIR from cfg
    if DATABASE_DIR.value is not None:
        database = DataBase(DATABASE_DIR.value)
        if dataset_id is None:
            Home()
        elif dataset_id not in set(database.datasets) | set(IN_MEMORY_DATASETS.keys()):
            solara.Error(f"Dataset {dataset_id} not found")
        else:
            with solara.Div(style={"margin-top": "-36px"}):
                # fix key to prevent rerender on opening dialogs
                # why is this needed? also doesnt work at the moment
                # seems like the dialogs dont play well with the grid layout
                DataSetView(dataset_id).key(f"dataset_view-{dataset_id}-{dark_effective}")
    else:
        solara.Error(
            "Database directory not configured. Please set database_dir in config or use --database-dir option."
        )


@solara.component  # type: ignore
def Page():
    router = solara.use_router()  # type: ignore
    if router.search is None:
        query_dict = {}
    else:
        query_dict = {k: v for k, v in parse_qsl(router.search)}

    dataset_id = query_dict.get("dataset_id", None)

    Main(dataset_id=cfg.dataset_id or dataset_id)


# %%



# %%

from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
import polars as pl
import ultraplot as uplt
from hdxms_datasets import DataBase
from ipymolstar import PDBeMolstar

from instagibbs.methods import (
    delta_dataframes,
    weighted_average,
)
from instagibbs.plot import (
    pdbemolstar_colors,
    pdbemolstar_tooltips,
    chiclet_quantitative,
    chiclet_categorical,
)
from instagibbs.plot.mpl import mpl_peptides
from instagibbs.quantities import RFU, DRFU
from instagibbs.preprocess import load_hdx_dataset
from hdxms_datasets.utils import slice_exposure
# %%

root = Path(__file__).parent.parent
database_dir = root / "tests" / "test_data"

# %%


def wt_avg_dataframe(df: pl.DataFrame, extend=True) -> pl.DataFrame:
    """Compute weighted average of a peptide value dataframe

    Args:
        df: Dataframe with peptide values
        extend: Whether to extend the dataframe to residue number zero. Defaults to True.

    Returns:
        Dataframe with weighted average values

    """
    dfs = [weighted_average(df, value=col).rename({"value": col}) for col in df.columns[2:]]
    combined = pl.concat(dfs, how="align")
    if extend:
        r_max = int(combined["r_number"].max())  # type: ignore
        extended = pl.DataFrame({"r_number": np.arange(0, r_max + 1)}).join(
            combined, on="r_number", how="left"
        )
        return extended

    return combined


# %%

DATASET = "HDX_D9096080"  # SecBs state data
DATASET = "HDX_5756A4C0"
PLOT_QTY = "frac_fd_control"
PLOT_QTY = "frac_max_uptake"


# %%
db = DataBase(database_dir)
dataset = db.load_dataset(DATASET)  # Should load the dataset from the database
states = load_hdx_dataset(dataset)

# %%
state_base = list(states.values())[0]  # take the first state as base
state_diff = list(states.values())[1]  # take the second state as diff

exposure = state_base.exposure.tolist()[0]

# %%
rfu_peptides = slice_exposure(state_base.data)[0]
# %%
fig, ax = plt.subplots()
annotate = {"fontsize": 10}
patches = mpl_peptides(
    ax,
    rfu_peptides,
    RFU,
    value=PLOT_QTY,
    annotate=annotate,
)
fig.tight_layout()
ax.set_title(f"RFU ({PLOT_QTY})")


# %%
delta_rfu_peptides = delta_dataframes(state_diff.data, state_base.data, value=PLOT_QTY, how="right")

fig, ax = plt.subplots()
annotate = {"fontsize": 10}
patches = mpl_peptides(
    ax,
    slice_exposure(delta_rfu_peptides)[0],
    DRFU,
    value=f"delta_{PLOT_QTY}",
    annotate=annotate,
)
ax.set_title("ΔRFU (mutant - WT)")
fig.tight_layout()

# %%
state_base.rfu[:, 2:].mean()


# %%
df = state_base.data.pivot(on="exposure", index=["start", "end"], values=PLOT_QTY)
rfu_residues = wt_avg_dataframe(df)
fig, ax = uplt.subplots(refwidth="100mm", refheight="40mm", sharex=False, grid=False)

# quantitative chiclet plot of RFU's
# the chiclet edges are linearly interpolated between y values
# TODO accept qty?
mesh = chiclet_quantitative(
    ax,
    rfu_residues,
    y_values=state_base.exposure,
    cmap=RFU.cmap,
    norm=RFU.norm,
    yscale="log",
    invert_y=False,
    edge_lines={"color": "grey"},
)
ax.colorbar(mesh, width="2.5mm", label="RFU")
ax.format(ylabel="Exposure time (s)", yformatter="log")

# %%
# categorical chiclet plot of RFU's
# each chiclet has the same width on the y axis
fig, ax = uplt.subplots(refwidth="100mm", refheight="40mm", sharex=False, grid=False)
mesh = chiclet_categorical(
    ax,
    rfu_residues,
    x="r_number",
    cmap=RFU.cmap,
    norm=RFU.norm,
    labels=lambda x: f"{float(x):.2f} s",
    invert_y=False,
    edge_lines={"color": "r"},
)
ax.colorbar(mesh, width="2.5mm", label="RFU")

# %%
drfu_peptides_wide = delta_dataframes(
    state_diff.data, state_base.data, value=PLOT_QTY, how="inner"
).pivot(on="exposure", index=["start", "end"], values=f"delta_{PLOT_QTY}")
# %%

drfu_residues_wide = wt_avg_dataframe(drfu_peptides_wide, extend=True)

# %%
fig, ax = uplt.subplots(refwidth="100mm", refheight="40mm", sharex=False, grid=False)
mesh = chiclet_quantitative(
    ax,
    drfu_residues_wide,
    y_values=state_base.exposure,
    cmap=DRFU.cmap,
    norm=DRFU.norm,
    yscale="log",
    invert_y=False,
    edge_lines={"color": "k"},
)
ax.colorbar(mesh, width="2.5mm", label="ΔRFU")
ax.format(ylabel="Exposure time (s)", yformatter="log")

# %%
fig, ax = uplt.subplots(refwidth="100mm", refheight="40mm", sharex=False, grid=False)
mesh = chiclet_categorical(
    ax,
    drfu_residues_wide,
    x="r_number",
    cmap=DRFU.cmap,
    norm=DRFU.norm,
    labels=lambda x: f"{float(x):.2f} s",
    invert_y=False,
    edge_lines={"color": "k"},
)
ax.colorbar(mesh, width="2.5mm", label="ΔRFU")

# %%

drfu_peptides_long = delta_dataframes(state_diff.data, state_base.data, value=PLOT_QTY, how="inner")
drfu_peptides_long

# %%
df_p = slice_exposure(drfu_peptides_long)[0].rename({f"delta_{PLOT_QTY}": "value"})
df_r = weighted_average(df_p)

# %% view delta RFU on a structure
value = str(state_base.exposure[0])
color_data = pdbemolstar_colors(df_r, DRFU)
tooltips = pdbemolstar_tooltips(df_r, DRFU)
view = PDBeMolstar(
    custom_data=dataset.structure.pdbemolstar_custom_data(),
    hide_water=True,
    color_data=color_data,
    tooltips=tooltips,
)

view

# %%


def get_peptide(df, start: int, end: int):
    return df.filter((pl.col("start") == start) & (pl.col("end") == end))


# %%
start, end = state_base.peptides.row(1)
start, end
# %%
peptide_base = get_peptide(state_base.data, start, end)
peptide_diff = get_peptide(state_diff.data, start, end)
peptide_base
# %%


# %%
peptide_base
# %%
d_uptake_corrected = (pl.col("uptake") / pl.col("fd_uptake") * pl.col("max_uptake")).alias(
    "d_uptake_corrected"
)
peptide_base = peptide_base.with_columns(d_uptake_corrected)
peptide_diff = peptide_diff.with_columns(d_uptake_corrected)

# %%
fig, ax = plt.subplots()

x = peptide_base["exposure"]
y = peptide_base["d_uptake_corrected"]
y1 = peptide_diff["d_uptake_corrected"]
ax.fill_between(x, y, y1, color="grey", alpha=0.5)

ax.scatter(peptide_base["exposure"], peptide_base["d_uptake_corrected"], label="WT")
ax.scatter(peptide_diff["exposure"], peptide_diff["d_uptake_corrected"], label="mutant")
max_uptake = peptide_base["max_uptake"].unique().item()
ax.axhline(max_uptake, color="gray", linestyle="--", label="max uptake")
ax.legend()
ax.set_ylim(0, None)

ax.set_xscale("log")
ax.set_xlabel("Exposure time (s)")
ax.set_ylabel("d-uptake (corrected)")
plt.show()

# %%



"""
Calculating error propagation from uptake sd to deltaG values
"""

# %%
from pathlib import Path

import numdifftools as nd
import numpy as np
import ultraplot as uplt
from hdxms_datasets import DataBase
from scipy.constants import R
from scipy.optimize import root_scalar
from uncertainties import umath
from uncertainties import unumpy as unp
import polars as pl
from numpy.typing import NDArray

from instagibbs.preprocess import load_hdx_dataset

# %%
root = Path(__file__).parent.parent
database_dir = root / "tests" / "test_data"
vault = DataBase(database_dir=database_dir)

# %%

DATASET = "HDX_0A55672B"
ds = vault.load_dataset(DATASET)

states = load_hdx_dataset(ds)
hdx_state = states[ds.get_state(0).name]

# %%

# get uptake and fd_uptake as uarrays (with uncertainties)
uptake = hdx_state.to_uarray("uptake")
fd_uptake = hdx_state.to_uarray("fd_uptake")
max_uptake = hdx_state.to_numpy("max_uptake")

uptake_corrected = uptake * (max_uptake / fd_uptake)


# %%
# Calculate areas (and errors) using trapezoidal rule on log t axis

t = np.log(hdx_state.exposure)
u_area = np.trapezoid(uptake_corrected, t, axis=1)

rel_errors = unp.std_devs(u_area) / unp.nominal_values(u_area)

# %%

# assume a flat 2 second error on exposure time; error for this datasets increases ~2fold
u_exposure = unp.uarray(hdx_state.exposure, 2.0 * np.ones_like(hdx_state.exposure))
u_t = np.array([umath.log(x) for x in u_exposure])

areas_t = np.trapezoid(uptake_corrected, u_t, axis=1)
rel_errors_t = unp.std_devs(areas_t) / unp.nominal_values(areas_t)

fig, ax = uplt.subplots()
ax.hist(rel_errors, bins="fd")
ax.hist(rel_errors_t, bins="fd", label="with uncertainties", alpha=0.5)
rel_errors_t.mean(), rel_errors.mean()
# %%
# comment out to use areas with t-error
# u_area = areas_t

# %%

# k_int geomean per peptide
k_int = np.empty(len(hdx_state.peptides), dtype=np.float64)
for i, (start, end) in enumerate(hdx_state.peptides.iter_rows()):
    pep_k_int = hdx_state.k_int[start - hdx_state.n_term : end - hdx_state.n_term + 1]
    nonzero = pep_k_int[np.nonzero(pep_k_int)]
    k_geomean = np.exp(np.mean(np.log(nonzero)))
    k_int[i] = k_geomean


# %%


def k_int_geomean(
    peptides: pl.DataFrame, k_int: NDArray[np.float64], n_term: int = 1
) -> NDArray[np.float64]:
    """Calculate the geometric mean of k_int values per peptide."""
    k_int_peptides = np.empty(len(peptides), dtype=np.float64)
    for i, (start, end) in enumerate(peptides.iter_rows()):
        pep_k_int = k_int[start - n_term : end - n_term + 1]
        nonzero = pep_k_int[np.nonzero(pep_k_int)]
        k_geomean = np.exp(np.mean(np.log(nonzero)))
        k_int_peptides[i] = k_geomean
    return k_int_peptides


max_d = max_uptake[:, 0]
area = unp.nominal_values(u_area)

# %%


def rootfunc_scalar(
    dG: float, time: np.ndarray, area: float, max_d: int, k: float, T: float
) -> float:
    d_ll = max_d * (1 - np.exp((-k * time) / (np.exp(-dG / (R * T)))))
    area_ll = np.trapezoid(d_ll, np.log(time))
    return area_ll - area


# %%
nboot = 50

bracket: tuple[float, float] = (20e3, -80e3)
t_lower, t_upper = np.log([hdx_state.exposure[0], hdx_state.exposure[-1]])
time = np.logspace(t_lower, t_upper, 250, endpoint=True, base=np.e)

dG_root = np.empty(len(hdx_state.peptides), dtype=np.float64)
dG_sd = np.empty(len(hdx_state.peptides), dtype=np.float64)
dG_sd_boot = np.empty(len(hdx_state.peptides), dtype=np.float64)
for pep_idx in range(len(hdx_state.peptides)):
    # Define the root function
    kwargs = {
        "time": time,
        "area": area[pep_idx],
        "max_d": max_d[pep_idx],
        "k": k_int[pep_idx],
        "T": hdx_state.temperature,
    }

    try:
        # Solve for dG
        ans = root_scalar(rootfunc_scalar, args=tuple(kwargs.values()), bracket=bracket)
        dG_root[pep_idx] = ans.root if ans.converged else np.nan

        d = nd.Derivative(rootfunc_scalar)(ans.root, **kwargs)
        dG_sd[pep_idx] = np.abs(1 / d) * unp.std_devs(u_area[pep_idx])

    except ValueError:
        dG_root[pep_idx] = np.nan

    # try bootstrapping errors
    if dG_root[pep_idx] is not np.nan:
        root_bootstrap = np.empty(nboot)
        for i in range(nboot):
            area_boot = np.random.normal(u_area[pep_idx].nominal_value, u_area[pep_idx].std_dev)
            boot_args = tuple((kwargs | {"area": area_boot}).values())
            try:
                ans = root_scalar(rootfunc_scalar, args=boot_args, bracket=bracket)
                root_bootstrap[i] = ans.root if ans.converged else np.nan
            except ValueError:
                print(i, pep_idx, "failed")

        dG_sd_boot[pep_idx] = np.nanstd(root_bootstrap)


dG_root, dG_sd, dG_sd_boot


# %%
# errors by finite differences
def rootfunc_vector(
    dG: np.ndarray, time: np.ndarray, area: np.ndarray, max_d: np.ndarray, k: np.ndarray, T: float
) -> np.ndarray:
    d_ll = max_d * (1 - np.exp((-np.outer(time, k)) / np.exp(-dG / (R * T))))

    area_ll = np.trapezoid(d_ll, np.log(time), axis=0)
    ans = area_ll - area
    ans[np.isnan(ans)] = 0.0  # Handle NaNs

    return ans  # Shape: (n,)


rootfunc_vector(dG_root, time, area, max_d, k_int, hdx_state.temperature)
args = (time, area, max_d, k_int, hdx_state.temperature)

eps = 1e-6
df_dg = (rootfunc_vector(dG_root + eps, *args) - rootfunc_vector(dG_root - eps, *args)) / (2 * eps)
dG_sd_fin_diff = np.abs(1 / df_dg) * unp.std_devs(u_area)
dG_sd_fin_diff
# %%

fig, ax = uplt.subplots()

ax.scatter(dG_sd / dG_root, label="numdifftools")
ax.scatter(dG_sd_boot / dG_root, label="bootstrap")
ax.scatter(dG_sd_fin_diff / dG_root, label="finite differences")
ax.format(ylim=(0, -0.2))

# %%

rel_err1 = np.abs(dG_sd / dG_root)
rel_err2 = np.abs(dG_sd_boot / dG_root)
rel_err3 = np.abs(dG_sd_fin_diff / dG_root)

n = len(rel_err1)
x = np.arange(n)

rng = np.random.default_rng(42)
jitter = 0.15

fig, axes = uplt.subplots(nrows=3, aspect=1.61, axwidth="100mm")
for ax in axes[:2]:
    ax.scatter(x - jitter, rel_err1, label="numdifftools", color="C0", alpha=0.8, s=40, marker="o")
    ax.scatter(x, rel_err2, label="bootstrap", color="C1", alpha=0.8, s=40, marker="s")
    ax.scatter(x + jitter, rel_err3, label="finite diff", color="C2", alpha=0.8, s=40, marker="^")
    ax.legend(loc="r", ncols=1)
    ax.format(
        xlabel="Peptide index",
        ylabel="Relative error",
        title="Relative error comparison of methods",
        grid=True,
    )

axes[0].format(ylim=(0, 0.2))
axes[1].format(ylim=(0, 0.02))

axes[2].scatter(dG_root)
axes[2].format(ylabel="dG (J/mol)", title="DeltaG vlaues")
uplt.show()
# %%
# Plot absolute errors
n = len(rel_err1)
x = np.arange(n)

rng = np.random.default_rng(42)
jitter = 0.15

fig, axes = uplt.subplots(nrows=3, aspect=2, axwidth="100mm")

axes[0].scatter(x - jitter, dG_sd, label="numdifftools", color="C0", alpha=0.8, s=40, marker="o")
axes[0].scatter(x, dG_sd_boot, label="bootstrap", color="C1", alpha=0.8, s=40, marker="s")
axes[0].scatter(
    x + jitter, dG_sd_fin_diff, label="finite diff", color="C2", alpha=0.8, s=40, marker="^"
)
axes[0].legend(loc="r", ncols=1)
axes[0].format(
    xlabel="Peptide index",
    ylabel="Absolute error",
    title="Absolute error comparison of methods",
    grid=True,
)

axes[0].format(ylim=(0, 2000))
axes[2].scatter(dG_root)
axes[2].format(ylabel="dG (J/mol)", title="DeltaG vlaues")
uplt.show()

# %%
fig, axes = uplt.subplots(nrows=4, aspect=2, axwidth="100mm")
axes[0].scatter(np.mean(unp.std_devs(uptake) / unp.nominal_values(uptake), axis=1))
axes[1].scatter(unp.std_devs(uptake) / unp.nominal_values(uptake))
axes[2].scatter(unp.std_devs(fd_uptake) / unp.nominal_values(fd_uptake))
axes[3].scatter(unp.std_devs(u_area) / unp.nominal_values(u_area))
fig.tight_layout()
uplt.show()

# %%



import warnings
from hdxms_datasets.models import DeuterationType, HDXDataSet, State
from hdxms_datasets.process import (
    compute_uptake_metrics,
    left_join,
    merge_peptide_tables,
    merge_peptides,
)
from hdxms_datasets.utils import get_peptides_by_type
from scipy.optimize import root_scalar
from hdxrate import k_int_from_sequence
from typing import Optional
import numpy as np
import polars as pl
import narwhals as nw
from instagibbs.models import HDXState as IGState


def has_fd_control(state: State) -> bool:
    """
    Check if the state has a fully deuterated control.
    """
    return any(
        peptide.deuteration_type == DeuterationType.fully_deuterated for peptide in state.peptides
    )


def rootfunc(t: float, k_int: np.ndarray, fd_measured: float) -> float:
    """
    Function to find the root for the back exchange time.
    """

    back_exchange = (1 - np.exp(-k_int * t)).sum()

    return len(k_int) - back_exchange - fd_measured


# TODO log back exchange conditions in datasets
# for now assume pH 4.5 and 0C
def try_correct_backexchange(
    sequence: str,
    fd_sequence: str,
    fd_uptake: float,
    pH: float = 4.5,
    temperature: float = 273.15,
    bracket=[0.1, 1e5],
) -> Optional[float]:
    # calculate the intrinsic rates for known fd control sequence
    k_int_fd = k_int_from_sequence(
        fd_sequence, pH_read=pH, temperature=temperature, ph_correction=False, exchange_type="HD"
    )

    sol = root_scalar(rootfunc, args=(k_int_fd, fd_uptake), bracket=bracket)
    if sol.converged:
        t_be = sol.root
        # calculate the intrinsic rates for the sequence with unknown back exchange
        k_int = k_int_from_sequence(
            sequence, pH_read=pH, temperature=temperature, ph_correction=False, exchange_type="HD"
        )

        # calculate the corrected FD uptake
        fd_uptake_corrected = len(k_int) - (1 - np.exp(-k_int * t_be)).sum().sum()
        return float(fd_uptake_corrected)
    else:
        return None


def adjust_fd_uptake_values(
    target_peptides: pl.DataFrame, fd_peptides: pl.DataFrame
) -> pl.DataFrame:
    """
    Correct the fully deuterated uptake values of target peptides based on known FD peptides.

    Given a known measurement of fully deuterated peptides, this function adjusts the uptake values
    of target peptides, matching by start and end values, but might differ in sequence due to mutations.

    Uptake values are corrected by first calculating an effective back exchange time for the FD peptides,
    then using this time to adjust the uptake values of the target peptides.

    """
    # group by peptide, take fd values
    agg = [pl.col("fd_uptake").unique().first(), pl.col("fd_uptake_sd").unique().first()]
    fd_values = (
        fd_peptides.group_by(["start", "end", "sequence"], maintain_order=True)
        .agg(agg)
        .drop_nulls("fd_uptake")
    )

    peptide_df = target_peptides[["start", "end", "sequence"]].unique(maintain_order=True)

    join = peptide_df.join(fd_values, on=["start", "end"], how="left", suffix="_fd")

    matched_peptides = join.filter(pl.col("sequence") == pl.col("sequence_fd"))
    mismatched_peptides = join.filter(pl.col("sequence") != pl.col("sequence_fd"))

    other = matched_peptides.drop("sequence_fd").rename(
        {"fd_uptake": "uptake", "fd_uptake_sd": "uptake_sd"}
    )

    if mismatched_peptides.is_empty():
        return other
    else:
        # for the mismatched peptides, we calculate the back exchange intrinsic rates for both sequences and use it to correct the FD uptake
        fd_uptake_values = []
        for row in mismatched_peptides.rows(named=True):
            value = try_correct_backexchange(row["sequence"], row["sequence_fd"], row["fd_uptake"])
            fd_uptake_values.append(value)

        # scale new sd with new uptake values
        uptake_sd = ((pl.col("uptake") / pl.col("fd_uptake")) * pl.col("fd_uptake_sd")).alias(
            "uptake_sd"
        )
        fixed = (
            mismatched_peptides.with_columns(pl.Series("uptake", fd_uptake_values))
            .with_columns(uptake_sd)
            .drop("sequence_fd", "fd_uptake_sd", "fd_uptake")
        )

        concat = pl.concat([other, fixed]).sort(by=["start", "end"])
        return concat


def process_missing_fd_control(target_state: State, fd_peptides: pl.DataFrame) -> pl.DataFrame:
    """
    fd_peptides dataframe should have fd_uptake, fd_uptake_sd columns

    """

    # TODO support multiple partially deuterated peptides (different pH, temperature)
    target_peptides = get_peptides_by_type(
        target_state.peptides, DeuterationType.partially_deuterated
    )
    assert target_peptides is not None
    target_peptides = (
        target_peptides.load().to_polars()["start", "end", "sequence"].unique(maintain_order=True)
    )
    corrected_fd_control = adjust_fd_uptake_values(target_peptides, fd_peptides)

    # process the rest of the state as usual
    pd_peptides = get_peptides_by_type(target_state.peptides, DeuterationType.partially_deuterated)
    assert pd_peptides is not None, "No partially deuterated peptides found in the dataset."
    nd_peptides = get_peptides_by_type(target_state.peptides, DeuterationType.non_deuterated)

    # merge with non-deuterated if available
    # TODO handle missing non-deuterated controls?
    if nd_peptides is None:
        output = pd_peptides.load()
    else:
        output = merge_peptide_tables(
            pd_peptides.load(),
            non_deuterated=nd_peptides.load(),
        )

    output = left_join(
        output, nw.from_native(corrected_fd_control), select_columns=["uptake"], prefix="fd"
    )

    # TODO take the compute step out of the function
    return compute_uptake_metrics(output, exception="ignore").to_polars()


def load_hdx_dataset(dataset: HDXDataSet) -> dict[str, IGState]:
    """
    Load HDX dataset and preprocess it.

    We process the whole dataset in batch to allow for sharing of FD controls between states.

    """
    processed_states: dict[str, pl.DataFrame] = {}
    states_with_fd = [s for s in dataset.states if has_fd_control(s)]
    for state in states_with_fd:
        merged = merge_peptides(state.peptides)
        processed_states[state.name] = compute_uptake_metrics(
            merged, exception="ignore"
        ).to_polars()

    # take the first processed state with FD control to derive FD control for the other states with it
    # or the one with the most peptides?
    # or we merge all FD states?  <- this #TODO
    fd_peptides_ref = next(iter(processed_states.values()))
    remaining_states = [s for s in dataset.states if not has_fd_control(s)]
    for target_state in remaining_states:
        result_df = process_missing_fd_control(target_state, fd_peptides_ref)
        processed_states[target_state.name] = result_df

    assert set(state.name for state in dataset.states) == set(processed_states.keys()), (
        "Not all states were processed."
    )

    out = {}
    for state in dataset.states:
        pd_peptides = get_peptides_by_type(state.peptides, DeuterationType.partially_deuterated)
        assert pd_peptides is not None, "No partially deuterated peptides found in the dataset."

        # check if all peptide mappings are equal, warn if not
        peptide_mappings = [p.structure_mapping for p in state.peptides]
        if not all(peptide_mappings[0] == mapping for mapping in peptide_mappings[1:]):
            warnings.warn("Peptide structure mappings are not all the same for state " + state.name)

        mapping = pd_peptides.structure_mapping

        metadata = {
            "pH": pd_peptides.pH,
            "temperature": pd_peptides.temperature,
            "d_percentage": pd_peptides.d_percentage,
            "n_term": state.protein_state.n_term,
            "c_term": state.protein_state.c_term,
            "sequence": state.protein_state.sequence,
        }

        df = processed_states[state.name]
        ig_state = IGState(
            df, metadata, structure=dataset.structure, structure_mapping=mapping, name=state.name
        )
        out[state.name] = ig_state

    # check if all states have the same structure mapping
    structure_mappings = [state.structure_mapping for state in out.values()]
    if not all(mapping == structure_mappings[0] for mapping in structure_mappings[1:]):
        warnings.warn("Not all states have the same structure mapping.")

    return out



from __future__ import annotations
import polars as pl
from collections import defaultdict

import numpy as np
import numpy.typing as npt
from instagibbs.utils import rgb_to_hex
from instagibbs.quantities import PlotQty
from hdxms_datasets.models import StructureMapping
import itertools


def augment_chain(
    data: list[dict],
    mapping: StructureMapping,
) -> list[dict]:
    """Augment a list of data with chain information"""

    chain_name = "auth_asym_id" if mapping.auth_chain_labels else "struct_asym_id"
    if mapping.chain:
        aug_data = []
        for elem, chain in itertools.product(data, mapping.chain):
            aug_data.append(elem | {chain_name: chain})
    else:
        aug_data = data

    return aug_data


def pdbemolstar_colors(
    df: pl.DataFrame,
    qty: PlotQty,
    mapping: StructureMapping = StructureMapping(),
) -> dict:
    """
    Generate a list of colors for residues in a DataFrame based on a colormap and normalization.

    Parameters:
        df: DataFrame containing residue numbers and values.
        qty: PlotQty instance containing colormap and normalization information.
        mapping:


    Returns:
        List of colors corresponding to the residue numbers.
    """

    # TODO make in sync with hdxms-datasets; currently duplicated code

    color_bytes = qty.get_colors(df["value"], bytes=True)
    color_hex = rgb_to_hex(color_bytes)  # type: ignore

    non_selected_color = qty.color_bad
    residue_name = "auth_residue_number" if mapping.auth_residue_numbers else "residue_number"

    data = [{residue_name: mapping.map(r), "color": c} for r, c in zip(df["r_number"], color_hex)]
    color_data = {
        "data": augment_chain(data, mapping),
        "nonSelectedColor": non_selected_color,
    }

    return color_data


def pdbemolstar_tooltips(
    df: pl.DataFrame,
    qty: PlotQty,
    mapping: StructureMapping = StructureMapping(),
    formatter: str = "{:.2f}",
) -> dict:
    """
    Create tooltip data for PDBe-MolStar visualization.
    This function generates a dictionary containing tooltip data for each residue,
    formatting values according to specified parameters.

    Parameters
    ----------
    df : pl.DataFrame
        DataFrame containing residue numbers and values.
    unit : str, default="kJ/mol"
        Unit to display after the value.
    qty : PlotQty
        PlotQty object
    formatter : str, default="{:.2f}"
        String format pattern for the value.


    Returns
    -------
    dict
        Dictionary with a "data" key containing a list of dictionaries with
        residue numbers and tooltip strings.

    """

    df = df.drop_nulls().drop_nans()

    residue_name = "auth_residue_number" if mapping.auth_residue_numbers else "residue_number"
    data = [
        {
            residue_name: mapping.map(int(resi)),
            "tooltip": qty.name + ": " + f"{formatter.format(val * qty.scale_factor)} {qty.unit}",
        }
        for resi, val in zip(df["r_number"], df["value"])
    ]

    tooltips = {
        "data": augment_chain(data, mapping),
    }

    return tooltips


def pymol_color_script(
    residues: list[int] | npt.NDArray[np.integer] | pl.Series,
    colors: list[str],
    no_coverage: str,
    residue_range: tuple[int, int],
) -> str:
    color_mapping = defaultdict(list)
    for r, c in zip(residues, colors):
        color_mapping[c.replace("#", "0x")].append(r)

    # any residue within the expected range that are not specified are
    # colored as no coverage
    missing_residues = set(range(*residue_range)) - set(residues)
    color_mapping[no_coverage.replace("#", "0x")] = list(missing_residues)

    s = ""
    for color, r_grp in color_mapping.items():
        s += f"color {color}, resi {'+'.join(map(str, r_grp))}\n"

    return s



import polars as pl
import numpy as np


# moved to hdxms-datasets
def find_wrap(
    peptides: pl.DataFrame,
    margin: int = 4,
    step: int = 5,
    wrap_limit: int = 200,
) -> int:
    """
    Find the minimum wrap value for a given list of intervals.

    Args:
        peptides: Dataframe with columns 'start' and 'end' representing intervals.
        margin: The margin applied to the wrap value. Defaults to 4.
        step: The increment step for the wrap value. Defaults to 5.
        wrap_limit: The maximum allowed wrap value. Defaults to 200.

    Returns:
        int: The minimum wrap value that does not overlap with any intervals.
    """
    wrap = step

    while True:
        peptides_y = peptides.with_columns(
            (pl.int_range(pl.len(), dtype=pl.UInt32).alias("y") % wrap)
        )

        no_overlaps = True
        for name, df in peptides_y.group_by("y", maintain_order=True):
            overlaps = (np.array(df["end"]) + 1 + margin)[:-1] >= np.array(df["start"])[1:]
            if np.any(overlaps):
                no_overlaps = False
                break
                # return wrap

        wrap += step
        if wrap > wrap_limit:
            return wrap_limit  # Return the maximum wrap limit if no valid wrap found
        elif no_overlaps:
            return wrap



# %%
from dataclasses import replace
from pathlib import Path

import matplotlib.pyplot as plt
import ultraplot as uplt
from hdxms_datasets.database import DataBase
from ipymolstar import PDBeMolstar

from instagibbs.methods import (
    ddG_from_area,
    dG_from_area,
    lasso_regression,
    ridge_regression,
    weighted_average,
)
from instagibbs.plot.alt import alt_peptides, alt_residues
from instagibbs.plot.mpl import mpl_peptides, scatter
from instagibbs.plot.structure import pdbemolstar_colors, pdbemolstar_tooltips
from instagibbs.preprocess import load_hdx_dataset
from instagibbs.quantities import DDG, DG

# %%
root = Path(__file__).parent.parent
database_dir = root / "tests" / "test_data"

# %%
DATASET = "HDX_0A55672B"  # SecBs state data
# DATASET = "HDX_BABA5B4C"  # Peterle 1MB

db = DataBase(database_dir)
dataset = db.load_dataset(DATASET)  # Should load the dataset from the database
# %%
states = load_hdx_dataset(dataset)

state_base = list(states.values())[0]  # take the first state as base
state_diff = list(states.values())[1]  # take the second state as diff

# %%
dG_peptides = dG_from_area(state_base)
ddG_peptides = ddG_from_area(state_base, state_diff)

# %%
dG_residues_wt_avg = weighted_average(dG_peptides)
dG_residues_ridge = ridge_regression(dG_peptides, alpha=0.1)

ddG_residues_wt_avg = weighted_average(ddG_peptides)
ddG_residues_lasso = lasso_regression(ddG_peptides, alpha=1.0)

# %%
alt_peptides(dG_peptides, DG).interactive()

# %%
alt_peptides(ddG_peptides, DDG).interactive()

# %%
alt_residues(
    ddG_residues_lasso,
    DDG,
    height=200,
    # value_sd='test',
)


# %%
fig, ax = uplt.subplots(refaspect=1.61, width="120mm")
cbar = {"width": "4mm", "pad": "1mm"}
rectangles = mpl_peptides(ax, dG_peptides, DG, colorbar=cbar)

# %%
fig, ax = plt.subplots()
annotate = {"fontsize": 10}
patches = mpl_peptides(
    ax,
    ddG_peptides,
    DDG,
    annotate=annotate,
)
fig.tight_layout()

# %%

fig, axes = uplt.subplots(nrows=2, sharex=True, axwidth="80mm", refaspect=1.6)
scatter(axes[0], dG_residues_wt_avg, DG)
scatter(axes[1], dG_residues_ridge, DG)

# %%

fig, axes = uplt.subplots(nrows=2, sharex=True, axwidth="80mm", refaspect=1.6)
scatter(axes[0], dG_residues_ridge, DG)
scatter(axes[1], ddG_residues_wt_avg, DDG)

# %%
# for the moment we assume all peptides in all states map to the structure in the same way
#

DDG_rescale = replace(DDG, vmin=-2, vmax=2)

# %%
mapping = state_base.structure_mapping
color_data = pdbemolstar_colors(ddG_residues_wt_avg, qty=DDG_rescale, mapping=mapping)
tooltips = pdbemolstar_tooltips(ddG_residues_wt_avg, qty=DDG_rescale, mapping=mapping)
color_data

# %%

view = PDBeMolstar(
    custom_data=dataset.structure.pdbemolstar_custom_data(),
    hide_water=True,
    color_data=color_data,
    tooltips=tooltips,
)
view

# %%



"""
Example script on how to directly subtract peptides and an associated value
(e.g., uptake) between two states in an HDX-MS dataset.

"""

# %%
from pathlib import Path

from hdxms_datasets.database import DataBase
from hdxms_datasets.utils import slice_exposure

import polars as pl
from hdxms_datasets.plot import plot_peptides
from instagibbs.methods import delta_dataframes

# %%
root = Path(__file__).parent.parent
database_dir = root / "tests" / "test_data"

# %%
DATASET = "HDX_0A55672B"

# %%
db = DataBase(database_dir)
dataset = db.load_dataset(DATASET)

p_left: pl.DataFrame = dataset.states[0].peptides[0].load().to_native()
p_right: pl.DataFrame = dataset.states[1].peptides[0].load().to_native()
# %%

delta_uptake = delta_dataframes(p_left, p_right, value="uptake")
delta_uptake.head()

# make slicer object to select exposure
data_slice = slice_exposure(delta_uptake)
data_slice[0]
# %%
# select first exposure and plot
# gray peptides have no matching peptide in right state
plot_peptides(data_slice[0], value="delta_uptake")



# %%
from __future__ import annotations

from collections import OrderedDict
from pathlib import Path
from typing import Any

import numpy as np
import polars as pl

base_v = np.vectorize(np.base_repr)


def rgb_to_hex(rgb_a: np.ndarray) -> np.ndarray:
    """Converts rgba
    input values are [0, 255]

    alpha is set to zero

    returns as '#000000'

    adapted from from pyhdx

    """

    if isinstance(rgb_a, np.ndarray):
        # todo: allow rgb arrays
        assert rgb_a.shape[-1] == 4
        if rgb_a.data.c_contiguous:
            rgba_array = rgb_a
        else:
            rgba_array = np.array(rgb_a)
    else:
        raise TypeError(f"Invalid type for 'rgb_a': {rgb_a}")

    ints = rgba_array.astype(np.uint8).view(dtype=np.uint32).byteswap()
    padded = np.char.rjust(base_v(ints // 2**8, 16), 6, "0")
    result = np.char.add("#", padded).squeeze()

    return result


def hex_to_rgb(hex_color: str) -> tuple[int, ...]:
    return tuple(int(hex_color.lstrip("#")[i : i + 2], 16) for i in (0, 2, 4))


def clean_types(d: Any) -> Any:
    """cleans up nested dict/list/tuple/other `d` for exporting as yaml/json

    Converts library specific types to python native types, including numpy dtypes,
    OrderedDict, numpy arrays

    # https://stackoverflow.com/questions/59605943/python-convert-types-in-deeply-nested-dictionary-or-array

    """
    if isinstance(d, np.floating):
        return float(d)

    if isinstance(d, np.integer):
        return int(d)

    if isinstance(d, np.ndarray):
        return d.tolist()

    if isinstance(d, list):
        return [clean_types(item) for item in d]

    if isinstance(d, tuple):
        return tuple(clean_types(item) for item in d)

    if isinstance(d, OrderedDict):
        return clean_types(dict(d))

    if isinstance(d, dict):
        return {k: clean_types(v) for k, v in d.items()}

    if isinstance(d, Path):
        return str(d)

    else:
        return d


def sort_columns(df: pl.DataFrame, drop_columns: list[str] = []) -> pl.DataFrame:
    """Sorts the columns of a DataFrame based on the order of the provided list."""
    # typing: mean might return nan/null
    cols_sorted = sorted(df.drop(drop_columns).columns, key=lambda x: df[x].drop_nans().mean())  # type: ignore

    return df.select(drop_columns + cols_sorted)


"""
def find_connected_components(A: np.ndarray) -> list[np.ndarray]:
    # Convert to sparse matrix format, for efficient processing
    A_sparse = csr_matrix(A)

    # Create an adjacency matrix that includes both rows and columns of A
    rows, cols = A_sparse.shape
    adj_matrix = vstack(
        [
            hstack([csr_matrix((rows, rows)), A_sparse]),
            hstack([A_sparse.T, csr_matrix((cols, cols))]),
        ]
    )

    # Find the connected components
    n_components, labels = connected_components(csgraph=adj_matrix, directed=False)

    # Initialize a list to hold the submatrices
    submatrices = []

    # Iterate over each component to extract the submatrix
    for i in range(n_components):
        # Find the row and column indices corresponding to the current component within the original matrix
        row_indices = np.where(labels[:rows] == i)[0]
        col_indices = np.where(labels[rows:] == i)[0]

        # Extract the block corresponding to the current component
        block = A[np.ix_(row_indices, col_indices)]

        # Append the block to the list of submatrices
        submatrices.append((row_indices, col_indices, block))

    return submatrices
"""



import os
from pathlib import Path
from typing import Optional

import click
from hdxms_datasets import load_dataset
from solara.__main__ import run

from instagibbs.config import (
    CONFIG_DEFAULT_DIR,
    CONFIG_HOME_DIR,
    update_config_from_yaml,
    cfg,
)
from instagibbs.web.constants import IN_MEMORY_DATASETS

try:
    from dotenv import load_dotenv

    load_dotenv()
except ImportError:
    pass

DEV_MODE = os.environ.get("INSTAGIBBS_DEBUG", "0") == "1"

ROOT = Path(__file__).parent
APP_PATH = ROOT / "web" / "app.py"


@click.group()
def cli():
    """Don't FRET! CLI for analyzing confocal solution smFRET data."""
    pass


def find_config_file(config_path: Path) -> Optional[Path]:
    if config_path.exists():
        return config_path
    elif (pth := CONFIG_HOME_DIR / config_path).exists():
        return pth
    elif (pth := CONFIG_DEFAULT_DIR / config_path).exists():
        return pth


def load_config(config_path: Path) -> None:
    resolved_cfg_path = find_config_file(Path(config_path))
    if not resolved_cfg_path:
        raise click.BadParameter(f"Configuration file '{config_path}' not found")

    update_config_from_yaml(resolved_cfg_path)
    click.echo("Loading config file at: " + str(resolved_cfg_path))


@cli.command(context_settings=dict(ignore_unknown_options=True, allow_extra_args=True))
@click.option(
    "--config",
    type=click.Path(file_okay=True, exists=True, dir_okay=False, path_type=Path),
    default=None,
    help="Configuration file to use",
)
@click.option(
    "--remote-url",
    help="URL for remote database",
)
@click.option(
    "--database-dir",
    type=click.Path(file_okay=False, exists=True, dir_okay=True, path_type=Path),
    help="Path to local database directory",
)
@click.option(
    "--zip-file",
    type=click.Path(file_okay=True, exists=True, dir_okay=False, path_type=Path),
    help="Path to .zip file containing a dataset to directly load",
)
@click.argument("solara_args", nargs=-1, type=click.UNPROCESSED)
def serve(
    config: Optional[Path] = None,
    remote_url: Optional[str] = None,
    database_dir: Optional[Path] = None,
    zip_file: Optional[Path] = None,
    solara_args=None,
):
    """Run the don't fret web application."""
    if config is not None:
        load_config(Path(config))

    # Update cfg with CLI options (CLI overrides config file)
    if remote_url is not None:
        cfg.remote_url = remote_url
    if database_dir is not None:
        cfg.database_dir = Path(database_dir)

    # if we have a .zip file, load it and add to in-memory datasets
    if zip_file is not None:
        dataset = load_dataset(zip_file)
        IN_MEMORY_DATASETS[dataset.hdx_id] = dataset

        # set the dataset_id to load it directly
        cfg.dataset_id = dataset.hdx_id

    solara_args = solara_args or tuple()
    if DEV_MODE and {"-a", "--auto-restart"}.isdisjoint(solara_args):
        solara_args = (*solara_args, "-a")
    elif "--production" not in solara_args:
        solara_args = (*solara_args, "--production")
    args = [str(APP_PATH), *solara_args]

    run(args)


@cli.command()
@click.option("--user", "user", is_flag=True, help="Create config file in user's home directory")
def config(user: bool):
    """Create a local or global default configuration file."""
    src = ROOT / "config" / "default.yaml"
    if user:
        (CONFIG_HOME_DIR).mkdir(exist_ok=True, parents=True)
        output = CONFIG_HOME_DIR / "instagibbs.yaml"
    else:
        output = Path.cwd() / "instagibbs.yaml"

    if output.exists():
        click.echo(f"Configuration file already exists at '{str(output)}'")
        return

    else:
        output.write_text(src.read_text())

    click.echo(f"Configuration file created at '{str(output)}'")


if __name__ == "__main__":
    cli()



import pytest
import polars as pl
from instagibbs.models import ensure_equal_exposure


@pytest.fixture
def sample_data() -> pl.DataFrame:
    """Create a sample DataFrame for testing"""
    return pl.DataFrame(
        {
            "start": [1, 1, 2, 2, 3, 3, 4, 4],
            "end": [10, 10, 15, 15, 20, 20, 25, 25],
            "exposure": [0.1, 0.5, 0.1, 0.5, 0.1, 0.5, 0.1, 0.3],
            "uptake": [0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9],
        }
    )


def test_ensure_equal_exposure(sample_data: pl.DataFrame):
    """Test the ensure_equal_exposure function"""

    result = ensure_equal_exposure(sample_data)
    assert result.is_empty()

    result = ensure_equal_exposure(sample_data, [0.1, 0.5])
    assert set(result["start"]) == {1, 2, 3}
    assert len(result) == 6

    result = ensure_equal_exposure(sample_data[:-1])
    assert set(result["start"]) == {1, 2, 3}
    assert len(result) == 6



# %%
from __future__ import annotations

from pathlib import Path

import numpy as np
import polars as pl
import pytest
from hdxms_datasets import DataBase, HDXDataSet

from instagibbs.methods import (
    ddG_from_area,
    dG_from_area,
    drfu,
    ridge_regression,
    weighted_average,
)
from instagibbs.models import HDXState
from instagibbs.preprocess import load_hdx_dataset

test_dir = Path(__file__).parent
database_dir = test_dir / "test_data"
vault = DataBase(database_dir=database_dir)

DATASET = "HDX_0A55672B"
STATE_WT = "Tetramer"


@pytest.fixture()
def dataset() -> HDXDataSet:
    ds = vault.load_dataset(DATASET)
    return ds


@pytest.fixture()
def states(dataset: HDXDataSet) -> dict[str, HDXState]:
    return load_hdx_dataset(dataset)


@pytest.fixture()
def reference_result() -> np.lib.npyio.NpzFile:
    return np.load(test_dir / "test_reference" / "secb_results.npz")


def test_load_state(dataset: HDXDataSet):
    states = load_hdx_dataset(dataset)
    hdx_state = states[STATE_WT]

    assert isinstance(hdx_state, HDXState)

    peptide = hdx_state.peptides[0]
    assert isinstance(peptide, pl.DataFrame)

    peptide_sequence = hdx_state.peptide_sequence
    assert len(peptide_sequence) == 63
    assert peptide_sequence[0] == "MTFQIQRIY"


def test_dG_from_area(states: dict[str, HDXState], reference_result: np.lib.npyio.NpzFile):
    hdx_state = states[STATE_WT]
    dG_peptides = dG_from_area(hdx_state)

    assert isinstance(dG_peptides, pl.DataFrame)
    assert len(dG_peptides) == 63

    # Compare with reference result
    ref_dG_peptides = reference_result["dG_peptides"]
    np.testing.assert_allclose(
        -dG_peptides["value"].to_numpy(), ref_dG_peptides, rtol=1e-5, equal_nan=True
    )

    dG_residue = weighted_average(dG_peptides)
    assert isinstance(dG_residue, pl.DataFrame)
    assert len(dG_residue) == 147

    # Compare with reference result
    ref_dG_residue = reference_result["dG_residue"]
    np.testing.assert_allclose(
        -dG_residue["value"].to_numpy(), ref_dG_residue, rtol=1e-5, equal_nan=True
    )

    dG_residue_ridge = ridge_regression(dG_peptides, alpha=0.1)
    assert isinstance(dG_residue_ridge, pl.DataFrame)
    assert len(dG_residue_ridge) == 147

    # Compare with reference result
    ref_dG_residue_ridge = reference_result["dG_residue_ridge"]
    test_dG_residue_ridge = dG_residue_ridge["value"].to_numpy()

    # sets NaNs from test to reference
    ref_dG_residue_ridge[np.isnan(test_dG_residue_ridge)] = np.nan

    np.testing.assert_allclose(
        -test_dG_residue_ridge, ref_dG_residue_ridge, rtol=1e-5, equal_nan=True
    )


def test_rfu(states: dict[str, HDXState], reference_result: np.lib.npyio.NpzFile):
    hdx_state = states[STATE_WT]
    rfu_df = hdx_state.rfu

    assert isinstance(rfu_df, pl.DataFrame)
    rfu_peptides = rfu_df[:, 2:].to_numpy()
    assert rfu_peptides.shape == (63, 5)

    # Compare with reference result
    ref_rfu_peptides = reference_result["rfu_peptides"]
    np.testing.assert_allclose(rfu_peptides, ref_rfu_peptides, rtol=1e-5)


def test_ddG_drfu(states: dict[str, HDXState], reference_result: np.lib.npyio.NpzFile):
    names = list(states.keys())
    base_state_name = names[0]
    diff_state_name = names[1]
    state_base = states[base_state_name]
    state_diff = states[diff_state_name]

    ddg_peptides = ddG_from_area(state_base, state_diff)
    assert isinstance(ddg_peptides, pl.DataFrame)

    assert len(ddg_peptides) == 63

    # Compare with reference result
    ref_ddG_peptides = reference_result["ddg_peptides"]
    np.testing.assert_allclose(
        ddg_peptides["value"].to_numpy(), ref_ddG_peptides, rtol=1e-5, equal_nan=True
    )

    drfu_df = drfu(state_base, state_diff)
    assert isinstance(drfu_df, pl.DataFrame)

    drfu_peptides = drfu_df.pivot(on="exposure", index=["start", "end"], values="drfu")
    drfu_peptides = drfu_peptides[:, 2:].to_numpy()

    assert drfu_peptides.shape == (63, 5)

    # Compare with reference result
    ref_drfu_peptides = reference_result["drfu_peptides"]
    np.testing.assert_allclose(drfu_peptides, ref_drfu_peptides, rtol=1e-5, equal_nan=True)



# %%
from __future__ import annotations

from hdxms_datasets.utils import slice_exposure
import matplotlib.pyplot as plt
from hdxms_datasets import DataBase
from pathlib import Path

from instagibbs.methods import drfu
from instagibbs.preprocess import load_hdx_dataset
from instagibbs.plot import mpl_peptides
from instagibbs.quantities import RFU, DRFU

# %%
root = Path(__file__).parent.parent
database_dir = root / "tests" / "test_data"

# %%

DATASET = "HDX_0A55672B"  # SecBs state data

db = DataBase(database_dir)
dataset = db.load_dataset(DATASET)  # Should load the dataset from the database
states = load_hdx_dataset(dataset)
# %%

state_base = list(states.values())[0]  # take the first state as base
state_diff = list(states.values())[1]  # take the second state as diff

data = slice_exposure(state_base.data)[0]
data

fig, ax = plt.subplots()
patches = mpl_peptides(ax, data, qty=RFU, value="frac_fd_control")

# %%
fig, ax = plt.subplots()
drfu_data = drfu(state_base, state_diff)
patches = mpl_peptides(ax, slice_exposure(drfu_data)[0], qty=DRFU, value="drfu")



import uuid
from itertools import product

import solara

DEFAULT_GRID_ELEMENT_HEIGHT = 10
DEFAULT_GRID_ELEMENT_WIDTH = 4


def new_card(view_type: str) -> dict:
    card_entry = {
        "id": uuid.uuid4(),
        "view_type": view_type,
    }
    return card_entry


def default_grid_layout() -> list[dict]:
    return [
        {
            "h": DEFAULT_GRID_ELEMENT_HEIGHT,
            "i": "0",
            "moved": False,
            "w": 14,
            "x": 0,
            "y": 0,
        },
        {
            "h": DEFAULT_GRID_ELEMENT_HEIGHT,
            "i": "1",
            "moved": False,
            "w": 14,
            "x": 0,
            "y": 12,
        },
        {
            "h": 2 * DEFAULT_GRID_ELEMENT_HEIGHT,
            "i": "2",
            "moved": False,
            "w": 10,
            "x": 7,
            "y": 0,
        },
    ]


def grid_elements(layout: dict) -> set:
    """Generate a set of grid elements based on the layout."""
    x_grid = range(layout["x"], layout["x"] + layout["w"])
    y_grid = range(layout["y"], layout["y"] + layout["h"])
    return set(product(x_grid, y_grid))


def find_free_grid_position(
    current_layout: list[dict],
    width=DEFAULT_GRID_ELEMENT_WIDTH,
    height=DEFAULT_GRID_ELEMENT_HEIGHT,
    grid_width: int = 12,
):
    existing_grid_elements = set()
    for layout in current_layout:
        existing_grid_elements.update(grid_elements(layout))

    # for x, y in product(range(grid_width - width + 1), range(100)):
    for y, x in product(range(100), range(grid_width - width + 1)):
        new_layout = {
            "h": height,
            "i": str(len(current_layout)),
            "moved": False,
            "w": width,
            "x": x,
            "y": y,
        }

        new_grid_elements = grid_elements(new_layout)

        if existing_grid_elements.isdisjoint(new_grid_elements):
            return new_layout
    else:
        return None


def td(*args):
    return solara.v.Html(tag="td", children=list(args))  # type: ignore


def th(title: str):
    return solara.v.Html(tag="th", children=[title])  # type: ignore


def tr():
    return solara.v.Html(tag="tr")  # type: ignore



# Adapted from: https://github.com/maresb/hatch-vcs-footgun-example
# Define the variable '__version__':
try:
    # If we are in an editable install, the _versioneer file
    # exist and we can use it to find the version
    from instagibbs._versioneer import get_versions

    # This will fail with LookupError if the package is not installed in
    # editable mode or if Git is not installed.
    __version__ = get_versions()["version"]
except ImportError:
    # If the project was build with hatch, there should be a _version.py file
    try:
        from instagibbs._version import __version__  # noqa: F401 # type: ignore
    except ModuleNotFoundError:
        # The user is probably trying to run this without having installed
        # the package, so complain.
        raise RuntimeError("instagibbs is not correctly installed. Please install it with pip.")



from __future__ import annotations

import os
from dataclasses import asdict, dataclass
from pathlib import Path

import dacite as dc
import yaml
from dacite.data import Data

from instagibbs.utils import clean_types

CONFIG_HOME_DIR = Path(os.getenv("XDG_CONFIG_HOME", Path.home() / ".config")) / "instagibbs"
CONFIG_DEFAULT_DIR = Path(__file__).parent


@dataclass
class Config:
    """Global configuration object"""

    database_dir: Path
    remote_url: str

    # when set, directly load this dataset, from memory, local database, or remote database
    dataset_id: str | None = None

    @classmethod
    def from_dict(cls, data: Data):
        config = dc.Config(type_hooks={Path: lambda v: Path(v).expanduser()})
        return dc.from_dict(cls, data, config)

    @classmethod
    def from_yaml(cls, fpath: Path):
        data = yaml.safe_load(fpath.read_text())
        return cls.from_dict(data)

    def to_yaml(self, fpath: Path) -> None:
        s = yaml.dump(clean_types(asdict(self)), sort_keys=False)
        fpath.write_text(s)

    def update(self, data: Data):
        new_data = {**self.__dict__, **data}

        # we use `from_dict` to cast to the correct types
        new_cfg = Config.from_dict(new_data)
        vars(self).update(vars(new_cfg))

    def copy(self) -> Config:
        return Config.from_dict(asdict(self))


def update_config_from_yaml(config_path: Path) -> Config:
    """Updates the global configuration object with settings from a YAML file."""

    data = yaml.safe_load(config_path.read_text())
    cfg.update(data)
    return cfg


cfg_file_paths = [
    Path().cwd() / "instagibbs.yaml",
    CONFIG_HOME_DIR / "instagibbs.yaml",
    CONFIG_DEFAULT_DIR / "default.yaml",
]

# take the first one which exists
cfg_fpath = next((p for p in cfg_file_paths if p.exists()), None)
assert cfg_fpath
cfg = Config.from_yaml(cfg_fpath)



from __future__ import annotations

from functools import cached_property

import numpy as np
import polars as pl
from hdxms_datasets.models import Structure, StructureMapping
from hdxrate import k_int_from_sequence
from polars.exceptions import ColumnNotFoundError
from uncertainties import unumpy as unp


def ensure_equal_exposure(data: pl.DataFrame, exposures: list[float] | None = None) -> pl.DataFrame:
    """Ensure that all peptides have the same exposure."""

    unique_exposures = exposures or data["exposure"].unique()
    cols = ["start", "end"]

    grps = data.group_by(cols)
    dfs = [df for name, df in grps if set(df["exposure"]) == set(unique_exposures)]

    if len(dfs) == 0:
        exp_filtered = pl.DataFrame()
    else:
        exp_filtered = pl.concat(dfs).sort(by=cols + ["exposure"])

    return exp_filtered


class HDXState:
    def __init__(
        self,
        peptide_data: pl.DataFrame,
        metadata: dict,
        structure: Structure,
        structure_mapping: StructureMapping = StructureMapping(),
        name: str = "",
    ):
        peptide_cols = ["start", "end"]

        data = peptide_data.drop_nans(["uptake", "fd_uptake"]).sort(peptide_cols + ["exposure"])

        uptake_pivot = data.pivot(
            on="exposure", index=["start", "end"], values="uptake"
        ).drop_nulls()

        # we keep fd_uptake nulls
        fd_uptake_pivot = data.pivot(on="exposure", index=["start", "end"], values="fd_uptake")

        # these are now a table of unique peptides
        self.peptides = uptake_pivot[peptide_cols]

        # take the unique peptides from the full input table and store as data
        self.data = data.join(self.peptides, on=peptide_cols, how="inner")

        self.peptide_sequence = self.peptides.join(
            self.data["start", "end", "sequence"].unique(), on=["start", "end"], how="left"
        )["sequence"]

        # take the d uptake values as numpy array (N_p, N_e)
        self.uptake = uptake_pivot[:, 2:].to_numpy()
        # fd uptake as numpy array (N_p, )
        self.fd_uptake = fd_uptake_pivot.join(self.peptides, on=peptide_cols, how="inner")[
            :, 2
        ].to_numpy()
        # Exposure as numpy array (N_e, )
        self.exposure = data["exposure"].unique(maintain_order=True).to_numpy()

        self.metadata = metadata
        self.structure = structure
        self.structure_mapping = structure_mapping
        self.name = name

    @property
    def temperature(self) -> float:
        return self.metadata["temperature"]

    @property
    def pH(self) -> float:
        return self.metadata["pH"]

    @property
    def d_percentage(self) -> float:
        return self.metadata["d_percentage"]

    @property
    def n_term(self) -> int:
        return self.metadata["n_term"]

    @property
    def c_term(self) -> int:
        return self.metadata["c_term"]

    @property
    def sequence(self) -> str:
        return self.metadata["sequence"]

    @property
    def intervals(self) -> list[tuple[int, int]]:
        """Return the intervals of the peptides as a list of tuples (start, end)"""
        return [(start, end + 1) for start, end in self.peptides.iter_rows()]

    @property
    def r_number(self) -> np.ndarray:
        return np.arange(self.peptides["start"][0], self.peptides["end"][-1] + 1)

    @property
    def max_uptake(self) -> np.ndarray:
        """Number of exchange-competent peptides"""
        # TODO this information is also in the dataframe (should be)
        max_uptake = np.array([sum([c != "P" for c in seq]) for seq in self.peptide_sequence])

        return max_uptake

    def pivot(self, value: str) -> pl.DataFrame:
        """Pivot the data on the given values column."""
        return self.data.pivot(on="exposure", index=["start", "end"], values=value)

    def to_numpy(self, value: str) -> np.ndarray:
        """Convert the specified values column to a numpy array."""
        return self.data.pivot(on="exposure", index=["start", "end"], values=value)[
            :, 2:
        ].to_numpy()

    def to_uarray(self, value: str) -> np.ndarray:
        """Convert the specified values column to a numpy array."""
        values = self.to_numpy(value)
        values_sd = self.to_numpy(value + "_sd")

        arr = unp.uarray(values, values_sd)
        return arr

    def try_uarray(self, value: str) -> np.ndarray:
        """Try to convert the specified values column to a numpy array with uncertainties."""
        try:
            return self.to_uarray(value)
        except (KeyError, ColumnNotFoundError):
            # if the value is not present, return a numpy array of NaNs
            return self.to_numpy(value)

    def get_rfu_peptides(self, exposure: float | str) -> pl.DataFrame:
        """Get the RFU (fraction with respect to FD values) for a specific exposure time.
        returns dataframe with columns: start, end, sequence, value

        """
        return (
            self.data.pivot(on="exposure", index=["start", "end"], values="frac_fd_control")
            .with_columns(self.peptide_sequence)
            .select(["start", "end", "sequence", str(exposure)])
            .rename({str(exposure): "value"})
        )

    @property
    def rfu(self) -> pl.DataFrame:
        return self.data.pivot(on="exposure", index=["start", "end"], values="frac_fd_control")

    @property
    def uptake_corrected(self) -> np.ndarray:
        """uptake normalized to measured max uptake
        this is the standard m / m0 correction
        returned array has object dtype with uncertainties (nominal values with standard deviations)
        if 'uptake' or 'fd_uptake' have uncertainties.
        """

        uptake = self.try_uarray("uptake")
        fd_uptake = self.try_uarray("fd_uptake")
        max_uptake = self.to_numpy("max_uptake")
        uptake_corrected = uptake * (max_uptake / fd_uptake)

        return uptake_corrected

    @cached_property
    def k_int(self) -> np.ndarray:
        """
        Returns an array of k_int values based on the given sequence, temperature, pH, and d_percentage.

        Returns:
            np.ndarray: An array of k_int values.
        """
        return k_int_from_sequence(
            self.sequence, self.temperature, self.pH, d_percentage=self.d_percentage
        )



from instagibbs.plot.helpers import find_wrap
from instagibbs.plot.mpl import chiclet_categorical, chiclet_quantitative, mpl_peptides
from instagibbs.plot.alt import alt_peptides, alt_rectangles
from instagibbs.plot.structure import pdbemolstar_colors, pdbemolstar_tooltips

__all__ = [
    "find_wrap",
    "pdbemolstar_colors",
    "pdbemolstar_tooltips",
    "mpl_peptides",
    "alt_peptides",
    "alt_rectangles",
    "chiclet_categorical",
    "chiclet_quantitative",
]



from __future__ import annotations

import warnings
from typing import Literal, Tuple
import polars as pl
import numpy as np
from scipy.constants import R
from scipy.optimize import root_scalar
from sklearn.linear_model import Lasso, LinearRegression, Ridge
from uncertainties import unumpy as unp

from instagibbs.models import HDXState


def subtract_peptides(
    peptides_left: pl.DataFrame,
    peptides_right: pl.DataFrame,
    value_left: str = "value",
    value_right: str = "value",
    value_out: str = "value",
):
    """
    Subtract the 'value' column of peptides_right from peptides_left.

    """

    placeholder_name = "_placeholder_"
    # Perform the subtraction
    # Rename the 'value' column in rfu_diff to avoid ambiguity after join
    peptides_right_renamed = peptides_right.rename({value_right: placeholder_name})

    # Perform a left join from rfu_base to rfu_diff_renamed
    # This keeps all rows from rfu_base and matching rows from rfu_diff_renamed
    # If no match, 'value_diff' will be null
    joined_df = peptides_left.join(
        peptides_right_renamed.select(["start", "end", placeholder_name]),
        on=["start", "end"],
        how="left",
    )

    if "sequence" in joined_df.columns:
        select = ["start", "end", "sequence", value_left]
    else:
        select = ["start", "end", value_left]

    # Subtract the values, nulls will propagate (value - null = null)
    subtracted_df = joined_df.with_columns(
        (pl.col(value_left) - pl.col(placeholder_name)).alias(value_out)
    ).select(select)

    return subtracted_df


def stack_uptake_curves(exposure: np.ndarray, uptake: np.ndarray) -> np.ndarray:
    Np, Ne = uptake.shape

    out = np.empty((Np, Ne, 2))
    out[:, :, 0] = exposure[np.newaxis, :]
    out[:, :, 1] = uptake

    return out


def k_int_geomean(peptides: pl.DataFrame, k_int: np.ndarray, n_term: int = 1) -> np.ndarray:
    """Calculate the geometric mean of k_int values per peptide."""
    k_int_peptides = np.empty(len(peptides), dtype=np.float64)
    for i, (start, end) in enumerate(peptides.iter_rows()):
        pep_k_int = k_int[start - n_term : end - n_term + 1]
        nonzero = pep_k_int[np.nonzero(pep_k_int)]
        k_geomean = np.exp(np.mean(np.log(nonzero)))
        k_int_peptides[i] = k_geomean
    return k_int_peptides


def rootfunc_scalar(
    dG: float, time: np.ndarray, area: float, max_d: int, k: float, T: float
) -> float:
    d_ll = max_d * (1 - np.exp((-k * time) / (np.exp(-dG / (R * T)))))
    area_ll = np.trapezoid(d_ll, np.log(time))
    return area_ll - area


def rootfunc_vector(
    dG: np.ndarray, time: np.ndarray, area: np.ndarray, max_d: np.ndarray, k: np.ndarray, T: float
) -> np.ndarray:
    d_ll = max_d * (1 - np.exp((-np.outer(time, k)) / np.exp(-dG / (R * T))))

    area_ll = np.trapezoid(d_ll, np.log(time), axis=0)
    ans = area_ll - area
    ans[np.isnan(ans)] = 0.0  # Handle NaNs

    return ans  # Shape: (n,)


def dG_from_area(
    hdx_state: HDXState,
    bracket: tuple[float, float] = (10e3, -50e3),
) -> pl.DataFrame:
    # TODO: allow errors on exposure time
    t = np.log(hdx_state.exposure)
    u_area = np.trapezoid(hdx_state.uptake_corrected, t, axis=1)
    k_int = k_int_geomean(hdx_state.peptides, hdx_state.k_int, hdx_state.n_term)
    max_d = hdx_state.to_numpy("max_uptake")[:, 0]
    area = unp.nominal_values(u_area)

    t_lower, t_upper = np.log([hdx_state.exposure[0], hdx_state.exposure[-1]])
    time = np.logspace(t_lower, t_upper, 250, endpoint=True, base=np.e)

    dG_root = np.empty(len(hdx_state.peptides), dtype=np.float64)
    for pep_idx in range(len(hdx_state.peptides)):
        args = (
            time,
            area[pep_idx],
            max_d[pep_idx],
            k_int[pep_idx],
            hdx_state.temperature,
        )

        try:
            # Solve for dG
            ans = root_scalar(rootfunc_scalar, args=args, bracket=bracket)
            dG_root[pep_idx] = ans.root if ans.converged else np.nan
        except ValueError:
            dG_root[pep_idx] = np.nan

    # Calculate finite difference errors
    eps = 1e-6
    args = (time, area, max_d, k_int, hdx_state.temperature)
    df_dg = (rootfunc_vector(dG_root + eps, *args) - rootfunc_vector(dG_root - eps, *args)) / (
        2 * eps
    )
    df_dg = np.where(df_dg == 0, np.nan, df_dg)  # Avoid division by zero
    dG_sd = np.abs(1 / df_dg) * unp.std_devs(u_area)

    return hdx_state.peptides.with_columns(
        hdx_state.peptide_sequence,
        pl.lit(dG_root).alias("value"),
        pl.lit(dG_sd).alias("value_sd"),
    )


def ddG_from_area(
    state_base: HDXState,
    state_diff: HDXState,
    intersect_exposure: Literal["interval", "identical"] = "identical",
) -> pl.DataFrame:
    if state_base.temperature != state_diff.temperature:
        warnings.warn("Comparing measurements with different temperatures")

    if len(state_diff.exposure) <= 1 or len(state_base.exposure) <= 1:
        raise ValueError("Must have more than 1 exposure time")

    r, t = intersect(state_base, state_diff, intersect_exposure=intersect_exposure)

    trapz_test = np.trapezoid(t.uptake_corrected, np.log10(t.exposure)[np.newaxis, :], axis=1)
    trapz_ref = np.trapezoid(r.uptake_corrected, np.log10(r.exposure)[np.newaxis, :], axis=1)

    trapz_area = trapz_test - trapz_ref
    ddG_sum = -np.log(10) * R * r.temperature * trapz_area
    length = r.peptides["end"] - r.peptides["start"] + 1
    ddG_mean = ddG_sum / length.to_numpy()

    ddG_df = r.peptides.with_columns(
        pl.lit(unp.nominal_values(ddG_mean)).alias("value"),
        pl.lit(unp.std_devs(ddG_mean)).alias("value_sd"),
    )
    peptides_sequence = state_base.peptides.with_columns(state_base.peptide_sequence)
    ddG_joined = peptides_sequence.join(ddG_df, on=["start", "end"], how="left")

    return ddG_joined


# mark for deprecation
def drfu(
    state_base: HDXState,
    state_diff: HDXState,
) -> pl.DataFrame:
    on = ["start", "end", "exposure"]
    value = "frac_fd_control"
    df_l = state_base.data[on + [value]]
    df_r = state_diff.data[on + [value]]

    joined = df_l.join(df_r, on=on, how="inner", suffix="_right")

    # Subtract the rfu values
    drfu = joined.with_columns(
        (pl.col("frac_fd_control_right") - pl.col("frac_fd_control")).alias("drfu")
    ).drop("frac_fd_control", "frac_fd_control_right")

    # Merge with original dataframe
    output = df_l[on].join(drfu, on=on, how="left")

    return output


# TODO potential target to move to hdxms_datasets ?
def delta_dataframes(
    df_left: pl.DataFrame,
    df_right: pl.DataFrame,
    value: str,
    on: list[str] | None = None,
    how: Literal["inner", "left", "right", "full", "semi", "anti", "cross"] = "right",
    prefix: str = "delta_",
) -> pl.DataFrame:
    """
    Calculate the difference between two DataFrames for a specific value column.
    Keeps all rows from df_left and matches rows from df_right based on the 'on' columns.

    Includes propagation of uncertainties if standard deviation columns are present.

    Args:
        df_left (pl.DataFrame): The left DataFrame.
        df_right (pl.DataFrame): The right DataFrame.
        value (str): The target column to calculate the difference for.
        how (str): The type of join to perform. Defaults to 'right'.
        on (list[str], optional): Columns to join on. Defaults to ["protein", "state", "start", "end", "exposure"],
            if present.
        prefix (str, optional): Prefix for the new columns. Defaults to "delta_".
    """

    on_columns = ["protein", "state", "start", "end", "exposure"]
    common_columns = set(df_left.columns).intersection(set(df_right.columns))
    if value not in common_columns:
        raise ValueError(f"Target column '{value}' must be present in both DataFrames.")

    if on is None:
        on = [col for col in on_columns if col in common_columns]

    select_columns = [value]
    if value + "_sd" in df_left.columns and value + "_sd" in df_right.columns:
        select_columns.append(value + "_sd")

    with_columns = [(pl.col(value) - pl.col(f"{value}_right")).alias(f"{prefix}{value}")]
    if value + "_sd" in select_columns:
        with_columns.append(
            (pl.col(f"{value}_sd") ** 2 + pl.col(f"{value}_sd_right") ** 2)
            .sqrt()
            .alias(f"{prefix}{value}_sd")
        )

    matching_peptides = df_left.join(df_right.select(on + select_columns), on=on, how=how)

    drop_columns = [f"{col}_right" for col in select_columns]
    output = matching_peptides.with_columns(with_columns).drop(drop_columns)

    return output


def coverage_blocks(intervals: list[tuple[int, int]]) -> Tuple[np.ndarray, np.ndarray]:
    """
    Convert a list of intervals into a matrix representation and calculate block sizes.
    Each block is defined as the regions between consecutive edges of the intervals, ie
    sections of residues which are covered by an unique set of peptides.

    Args:
        intervals: A list of (start, end) tuples representing intervals
    Returns:
        A tuple containing:
        - X: A matrix where rows correspond to blocks (regions between consecutive edges)
             and columns correspond to intervals. The value X[i, j] is the size of block i
             if it's covered by interval j, otherwise 0. The shape of X is (n_blocks, n_intervals).
        - block_sizes: An array of sizes for each block
    Example:
        >>> coverage_blocks([(1, 5), (3, 7)])
        (array([[2, 0],  # Block [1, 3] is only covered by first interval
                [2, 2],  # Block [3, 5] is covered by both intervals
                [0, 2]]), # Block [5, 7] is only covered by second interval
         array([2, 2, 2])) # Sizes of the blocks [1, 3], [3, 5], [5, 7]
    """
    edges = np.array(sorted({a for tup in intervals for a in tup}))
    block_sizes = np.diff(edges)

    X = np.zeros((len(edges) - 1, len(intervals)), dtype=int)
    for i, tup in enumerate(intervals):
        e_start, e_stop = np.searchsorted(edges, tup)
        blocks = np.diff(edges[e_start : e_stop + 1])

        X[e_start:e_stop, i] = blocks

    return X, block_sizes


def weighted_average(
    df: pl.DataFrame,
    value: str = "value",
) -> pl.DataFrame:
    """
    Calculate weighted average of values over residues based on peptide intervals.
    Any NaN values in the `values` column will be dropped before calculation.

    Args:
        df (pl.DataFrame): DataFrame containing peptide intervals and values.
        start (str): Column name for the start of the peptide interval.
        end (str): Column name for the end of the peptide interval (inclusive interval).
        values (str): Column name for the values to average.
    """

    expected_columns = {"start", "end", value}
    if not expected_columns.issubset(df.columns):
        raise ValueError(f"DataFrame must contain columns: {expected_columns}")

    df = df.drop_nans([value]).drop_nulls([value])

    peptide_intervals = [(_start, _end + 1) for _start, _end in zip(df["start"], df["end"])]
    r_number = np.arange(df["start"].min(), df["end"].max() + 1)  # type: ignore
    X, block_sizes = coverage_blocks(peptide_intervals)

    with np.errstate(invalid="ignore"):
        Z_norm = X / np.sum(X, axis=1, keepdims=True)
    value_blocks = Z_norm.dot(df[value])
    values_residue = np.repeat(value_blocks, block_sizes)

    return pl.DataFrame({"r_number": r_number, "value": values_residue})


def _linear_regression_base(
    df: pl.DataFrame,
    model,
    value: str = "value",
    value_sd: str | None = None,
    n_bootstrap: int = 50,
) -> pl.DataFrame:
    """Base function for linear regression methods (Ridge, Lasso, etc.)"""
    expected_columns = {"start", "end", value}
    if not expected_columns.issubset(df.columns):
        raise ValueError(f"DataFrame must contain columns: {expected_columns}")

    df = df.drop_nans([value]).drop_nulls([value])

    peptide_intervals = [(_start, _end + 1) for _start, _end in zip(df["start"], df["end"])]
    r_number = np.arange(df["start"].min(), df["end"].max() + 1)  # type: ignore

    X, block_sizes = coverage_blocks(peptide_intervals)
    X_norm = X / np.sum(X, axis=0)

    y = df[value].to_numpy()
    model.fit(X_norm.T, y)
    values_residue = np.repeat(model.coef_ + model.intercept_, block_sizes)
    no_coverage = np.repeat(X.sum(axis=1) == 0, block_sizes)
    values_residue[no_coverage] = np.nan
    data_dict = {"r_number": r_number, "value": values_residue}

    value_sd = f"{value}_sd" if value_sd is None else value_sd
    if value_sd in df.columns and n_bootstrap != 0:
        y_std = df[value_sd].to_numpy()
        out = []
        for _ in range(n_bootstrap):
            y_boot = np.random.normal(y, y_std)
            model.fit(X_norm.T, y_boot)
            out.append(model.coef_ + model.intercept_)

        errors = np.std(np.array(out), axis=0)
        sd_values_residue = np.repeat(errors, block_sizes)
        sd_values_residue[no_coverage] = np.nan
        data_dict["value_sd"] = sd_values_residue

    return pl.DataFrame(data_dict)


def ridge_regression(
    df: pl.DataFrame,
    value: str = "value",
    value_sd: str | None = None,
    alpha: float = 1.0,
    n_bootstrap: int = 50,
) -> pl.DataFrame:
    if alpha == 0:
        model = LinearRegression()
    else:
        model = Ridge(alpha=alpha)

    return _linear_regression_base(df, model, value, value_sd, n_bootstrap)


# works well for ddG since the data is already centered
# see: https://medium.com/@mukulranjan/how-does-lasso-regression-l1-encourage-zero-coefficients-but-not-the-l2-20e4893cba5d
def lasso_regression(
    df: pl.DataFrame,
    value: str = "value",
    value_sd: str | None = None,
    alpha: float = 1.0,
    n_bootstrap: int = 50,
) -> pl.DataFrame:
    if alpha == 0:
        model = LinearRegression()
    else:
        model = Lasso(alpha=alpha, fit_intercept=False)

    return _linear_regression_base(df, model, value, value_sd, n_bootstrap)


def intersect(
    s_left: HDXState,
    s_right: HDXState,
    intersect_exposure: Literal["interval", "identical"] = "identical",
) -> tuple[HDXState, HDXState]:
    """Intersect HDX states by their peptides and exposures.

    Args:
        s_left (HDXState): The first HDX state.
        s_right (HDXState): The second HDX state.
        intersect_exposure (Literal["interval", "identical"]): How to intersect exposures.
            - "identical": Only keep exposures that are identical in both states.
            - "interval": Selects the largest exposure interval in the set of matching exposures,
                  ie allows for missing exposure data at intermediate time points.
    Returns:
        tuple[HDXState, HDXState]: A tuple containing two HDXState objects with intersected data.
    """

    if intersect_exposure == "identical":
        on = ["start", "end", "exposure"]
        d_match_left = s_left.data.join(s_right.data[on], on=on, how="inner")
        d_match_right = s_right.data.join(s_left.data[on], on=on, how="inner")
    elif intersect_exposure == "interval":
        matching_exposure = set(s_left.exposure) & set(s_right.exposure)
        vmin, vmax = min(matching_exposure), max(matching_exposure)

        d_left = s_left.data.filter(pl.col("exposure").is_between(vmin, vmax))
        d_right = s_right.data.filter(pl.col("exposure").is_between(vmin, vmax))

        on = ["start", "end"]
        d_match_left = d_left.join(d_right[on], on=on, how="inner")
        d_match_right = d_right.join(d_left[on], on=on, how="inner")

    s_match_left = HDXState(d_match_left, s_left.metadata, structure=s_left.structure)
    s_match_right = HDXState(d_match_right, s_right.metadata, structure=s_right.structure)

    return s_match_left, s_match_right


# def intersect_peptides(p_left: Protein, p_right: Protein) -> tuple[Protein, Protein]:
#     matching_peptides = set(p_left.peptides) & set(p_right.peptides)

#     bools_left = np.array([peptide in matching_peptides for peptide in p_left.peptides])
#     bools_right = np.array([peptide in matching_peptides for peptide in p_right.peptides])

#     return p_left.index(bools_left, axis=0), p_right.index(bools_right, axis=0)


# def intersect_exposures(
#     p_left: Protein, p_right: Protein, intersect: Literal["interval", "identical"] = "identical"
# ) -> tuple[Protein, Protein]:
#     matching_exposure = np.intersect1d(p_left.exposure, p_right.exposure)

#     if intersect == "identical":
#         bools_left = np.isin(p_left.exposure, matching_exposure)
#         bools_right = np.isin(p_right.exposure, matching_exposure)
#     elif intersect == "interval":
#         vmin, vmax = np.min(matching_exposure), np.max(matching_exposure)
#         bools_left = (p_left.exposure >= vmin) & (p_left.exposure <= vmax)
#         bools_right = (p_right.exposure >= vmin) & (p_right.exposure <= vmax)
#     else:
#         raise ValueError("`intersect` must be 'identical' or 'interval'")

#     return p_left.index(bools_left, axis=1), p_right.index(bools_right, axis=1)


def D_t(dG, time, temperature, k_int):
    """Calculate D-uptake as a function of time"""
    D_t = 1 - np.exp(
        np.divide(
            -k_int * time,
            1 + np.exp(dG / (R * temperature)),
        )
    )

    return D_t


def func_root_dg(
    dG: float, time: float, d_uptake: float, temperature: float, k_int: np.ndarray
) -> float:
    """function find root for dG given d_uptake and time"""
    dt = D_t(dG, time, temperature, k_int)

    return np.sum(dt) - d_uptake


def func_root_time(
    time: float, dG: float, d_uptake: float, temperature: float, k_int: np.ndarray
) -> float:
    """function find root for time given dG and d_uptake"""
    dt = D_t(dG, time, temperature, k_int)

    return np.sum(dt) - d_uptake



from .config import (
    CONFIG_DEFAULT_DIR,
    CONFIG_HOME_DIR,
    Config,
    cfg,
    update_config_from_yaml,
)

__all__ = [
    "CONFIG_DEFAULT_DIR",
    "CONFIG_HOME_DIR",
    "Config",
    "cfg",
    "update_config_from_yaml",
]



"""
Load a dataset into the web interface.
to run:
`solara run templates/load_dataset_web_interface.py`

"""

# %%
from pathlib import Path
from hdxms_datasets.database import DataBase
import solara
from instagibbs.web.app import DATABASE_DIR, Main

# %%

root = Path(__file__).parent.parent
database_dir = root / "tests" / "test_data"
DATABASE_DIR.set(database_dir)
DATASET = "HDX_0A55672B"

database = DataBase(DATABASE_DIR.value)
print(database.datasets)
# %%


@solara.component  # type: ignore
def Page():
    if DATABASE_DIR.value != database_dir:
        DATABASE_DIR.set(database_dir)

    Main(DATASET)


# %%



import polars as pl

from instagibbs.plot.helpers import find_wrap
import altair as alt
from instagibbs.quantities import PlotQty
from uncertainties import ufloat


def alt_rectangles(peptides: pl.DataFrame, wrap: int | None = None) -> pl.DataFrame:
    wrap = find_wrap(peptides, step=1) if wrap is None else wrap
    columns = [
        (pl.col("start") - 0.5).alias("x"),
        (pl.col("end") + 0.5).alias("x2"),
        (wrap - (pl.col("idx") % wrap)).alias("y"),
    ]

    rectangles = (
        peptides["start", "end"]
        .with_row_index("idx")
        .with_columns(columns)
        .with_columns((pl.col("y") - 1).alias("y2"))
    )

    return rectangles


def alt_peptides(
    peptides: pl.DataFrame,
    qty: PlotQty,
    value: str = "value",
    value_sd: str | None = None,
    width: str | int = "container",
    height: str | int = 350,
    wrap: int | None = None,
    fill_nan: bool = True,
) -> alt.Chart:
    if fill_nan:
        # nan values can cause problems in serialization
        peptides = peptides.fill_nan(None)

    value_sd = value_sd or f"{value}_sd"

    if value_sd in peptides.columns:
        tooltip_value = []
        for v, v_sd in zip(peptides[value], peptides[value_sd]):
            if v is not None and v_sd is not None:
                z = ufloat(v, v_sd).std_score(0)
                tooltip_value.append(
                    f"{v * qty.scale_factor:.2f} \u00b1 {v_sd * qty.scale_factor:.2f} ({z:.1f}\u03c3)"  # type: ignore
                )
            else:
                tooltip_value.append("NaN")
    else:
        tooltip_value = [
            f"{value * qty.scale_factor:.2f}" if value is not None else ""
            for value in peptides[value]
        ]

    rectangles = alt_rectangles(peptides, wrap=wrap)
    peptide_source = peptides.join(rectangles, on=["start", "end"], how="left").with_columns(
        pl.col(value) * qty.scale_factor, pl.Series(tooltip_value).alias("tooltip_value")
    )

    invalid = {"color": {"value": qty.color_bad}}
    peptide_chart = (
        alt.Chart(peptide_source)
        .mark_rect(
            stroke="black",
        )
        .encode(
            x=alt.X("x:Q", title="Residue Number"),
            y=alt.Y("y:Q", title="", axis=alt.Axis(ticks=False, domain=False, labels=False)),
            x2=alt.X2("x2:Q"),
            y2=alt.Y2("y2:Q"),
            tooltip=[
                alt.Tooltip("idx:Q", title="Index"),
                alt.Tooltip("start:Q", title="Start"),
                alt.Tooltip("end:Q", title="End"),
                alt.Tooltip("sequence:N", title="Sequence"),
                alt.Tooltip("tooltip_value:N", title=qty.label),
            ],
            color=alt.Color(f"{value}:Q", scale=qty.alt_scale, title=qty.label),
        )
        .configure_scale(invalid=invalid)
    )

    return peptide_chart.properties(height=height, width=width)


def alt_residues(
    residues: pl.DataFrame,
    qty: PlotQty,
    value: str = "value",
    value_sd: str | None = None,
    r_number: str = "r_number",
    width: str | int = "container",
    height: str | int = 350,
    pad: float | None = None,
    fill_nan: bool = True,
) -> alt.LayerChart | alt.Chart:
    if fill_nan:
        # nan values can cause problems in serialization
        residues = residues.fill_nan(None)

    zoom = alt.selection_interval(bind="scales")

    value_sd = f"{value}_sd" if value_sd is None else value_sd
    residue_source = residues.with_columns(pl.col(value) * qty.scale_factor)
    if pad is None:
        scale = alt.Undefined
    else:
        rv_min = residue_source[value].min()
        rv_max = residue_source[value].max()

        pad = (rv_max - rv_min) * 0.1  # type: ignore
        domain = rv_min - pad, rv_max + pad  # type: ignore

        scale = alt.Scale(domain=domain)  # type: ignore

    if value_sd in residues.columns:
        residue_source = residue_source.with_columns(pl.col(value_sd) * qty.scale_factor)
        residue_source = residue_source.with_columns(
            (pl.col(value) - pl.col(value_sd)).alias("y"),
            (pl.col(value) + pl.col(value_sd)).alias("y2"),
        )
        error_chart = (
            alt.Chart(residue_source)
            .mark_errorbar(ticks=True)
            .encode(
                alt.X(f"{r_number}:Q", title="Residue Number"),
                alt.Y("y:Q", title=qty.label, scale=scale),
                alt.Y2("y2:Q"),
                # color=alt.value(NO_COVERAGE),
            )
        )
    else:
        error_chart = None

    invalid = {"color": {"value": qty.color_bad}}
    residue_chart = (
        alt.Chart(residue_source)
        .mark_circle(size=100)
        .encode(
            x=alt.X(
                f"{r_number}:Q",
                title="Residue Number",
            ),
            y=alt.Y(f"{value}:Q", title=qty.label, scale=scale),
            tooltip=[
                alt.Tooltip(f"{r_number}:Q", title="Residue Number"),
                alt.Tooltip(f"{value}:Q", title=qty.label, format=".2f"),  # TODO: std tooltip
            ],
            color=alt.Color(f"{value}:Q", scale=qty.alt_scale, title=qty.label),
        )
        .add_params(zoom)
    )

    if error_chart is not None:
        chart = error_chart + residue_chart
    else:
        chart = residue_chart

    return chart.properties(height=height, width=width).configure_scale(invalid=invalid)



from dataclasses import dataclass, field, fields
from typing import Literal, TypeAlias

from hdxms_datasets.models import StructureMapping
import polars as pl
import solara

from instagibbs.methods import (
    ddG_from_area,
    dG_from_area,
    lasso_regression,
    ridge_regression,
    subtract_peptides,
    weighted_average,
)
from instagibbs.models import HDXState
from instagibbs.quantities import DDG, DG, DRFU, RFU, PlotQty


ReductionType: TypeAlias = Literal["weighted_average", "ridge_regression", "lasso_regression"]


@dataclass
class DataManager:
    hdx_states: dict[str, HDXState]
    # TODO as enum; with LUT for plotqtys
    qty: solara.Reactive[PlotQty] = field(default_factory=lambda: solara.reactive(DG))
    base_state: solara.Reactive[str] = field(init=False)
    diff_state: solara.Reactive[str] = field(init=False)

    exposure: solara.Reactive[str] = field(init=False)  # base/diff exposure? maybe not allow
    residue_reduction: solara.Reactive[ReductionType] = field(
        default_factory=lambda: solara.Reactive("weighted_average")
    )
    alpha: solara.Reactive[float] = field(default_factory=lambda: solara.reactive(0.1))

    _supported_qtys = {qty.name: qty for qty in [DG, DDG, RFU, DRFU]}

    def __post_init__(self):
        all_states = list(self.hdx_states.keys())
        self.base_state = solara.Reactive(all_states[0])
        self.diff_state = solara.Reactive(all_states[1] if len(all_states) > 1 else all_states[0])

        # TODO RFUs check for exposure in both states
        self.exposure = solara.Reactive(
            str(self.hdx_states[self.base_state.value].exposure.tolist()[0])
        )

    @property
    def hdx_base(self) -> HDXState:
        return self.hdx_states[self.base_state.value]

    @property
    def hdx_diff(self) -> HDXState:
        return self.hdx_states[self.diff_state.value]

    def get_peptide_df(self) -> pl.DataFrame:
        qty = self.qty.value
        if qty == DG:
            peptides = dG_from_area(self.hdx_base)
        elif qty == DDG:
            peptides = ddG_from_area(self.hdx_base, self.hdx_diff)
        elif qty == RFU:
            peptides = self.hdx_base.get_rfu_peptides(self.exposure.value)
        elif qty == DRFU:
            rfu_base = self.hdx_base.get_rfu_peptides(self.exposure.value)
            rfu_diff = self.hdx_diff.get_rfu_peptides(self.exposure.value)
            peptides = subtract_peptides(rfu_base, rfu_diff)
        else:
            raise ValueError(f"Unsupported quantity: {qty}")

        return peptides

    def get_residue_df(self) -> pl.DataFrame:
        peptides = self.get_peptide_df()

        if self.residue_reduction.value == "weighted_average":
            residues = weighted_average(peptides)
        elif self.residue_reduction.value == "ridge_regression":
            residues = ridge_regression(peptides, alpha=self.alpha.value)  # for DG/RFU
        elif self.residue_reduction.value == "lasso_regression":
            residues = lasso_regression(peptides, alpha=self.alpha.value)  # for DDG/DRFU
        else:
            raise ValueError(f"Unknown residue reduction method: {self.residue_reduction.value}")

        return residues

    def set_qty(self, qty_name: str):
        if qty_name not in self._supported_qtys:
            raise ValueError(f"Unsupported quantity: {qty_name}")
        new_qty = self._supported_qtys[qty_name]
        if new_qty in [DDG, DRFU] and self.residue_reduction.value == "ridge_regression":
            # lasso regression is not supported for DDG/DRFU
            self.residue_reduction.value = "weighted_average"
        elif new_qty in [DG, RFU] and self.residue_reduction.value == "lasso_regression":
            # ridge regression is not supported for DG/RFU
            self.residue_reduction.value = "weighted_average"

        self.qty.value = new_qty

    @property
    def qty_values(self) -> list[str]:
        """Return the names of the available quantities."""
        return list(self._supported_qtys.keys())

    @property
    def state_values(self) -> list[str]:
        """Return the names of the available states."""
        return list(self.hdx_states.keys())

    @property
    def exposure_values(self) -> list[str]:
        """Return the available exposure values from the base state."""
        # TODO ensure this always matches dataframe column names (or take them from the column names?)
        return [str(e) for e in self.hdx_base.exposure.tolist()]

    @property
    def pdbe_molstar_custom_data(self) -> dict:
        return self.hdx_base.structure.pdbemolstar_custom_data()

    @property
    def structure_mapping(self) -> StructureMapping:
        return self.hdx_base.structure_mapping

    @property
    def is_delta(self) -> bool:
        """Check if the current quantity is a delta quantity (DDG or DRFU)."""
        return self.qty.value in [DDG, DRFU]

    @property
    def label(self) -> str:
        """Return a label for the current quantity."""
        if self.is_delta:
            label = f"{self.qty.value.label}: {self.hdx_base.name} - {self.hdx_diff.name}"
        else:
            label = f"{self.qty.value.label}: {self.hdx_base.name}"

        return label

    def clone(self):
        field_names = [field_info.name for field_info in fields(self)]

        new_dm = DataManager(self.hdx_states)
        for field_name in field_names:
            field_self = getattr(self, field_name)
            if isinstance(field_self, solara.Reactive):
                field_other = getattr(new_dm, field_name)
                field_other.set(field_self.value)
        return new_dm



from __future__ import annotations
from matplotlib.collections import QuadMesh
import polars as pl
from typing import Callable, Iterable

import matplotlib.cm as cm
import numpy as np
from matplotlib.axes import Axes
from matplotlib.colorbar import Colorbar
from matplotlib.colors import Colormap, Normalize
from matplotlib.patches import Patch, Rectangle
from mpl_toolkits.axes_grid1.axes_divider import make_axes_locatable
from instagibbs.plot.helpers import find_wrap
from instagibbs.quantities import PlotQty
import ultraplot as uplt


def chiclet_quantitative(
    ax: uplt.Axes,
    df: pl.DataFrame,
    x: str = "r_number",
    y_values: Iterable[float] | Callable[[str], float] | None = None,
    cmap: Colormap | None = None,
    norm: Normalize | None = None,
    N: int = 256,
    yscale: str = "log",
    invert_y: bool = False,
    edge_lines: bool | dict = True,
) -> QuadMesh:
    """
    Plot a colormap of quantitative data as discrete chiclets.
    This function creates a colormap visualization where the y-axis represents discrete values
    (typiclly exposure time) and the x-axis represents positions (typically residue numbers).
    Each 'chiclet' or cell in the plot is colored according to the value at that position and y-value.
    Parameters
    ----------
    ax : uplt.Axes
        The axes on which to draw the plot.
    df : pl.DataFrame
        Input DataFrame where one column is the x coordinates, and remaining columns
        are the data values for different y levels.
    x : str, default="r_number"
        Column name in the DataFrame to use as x-coordinates.
    y_values : Iterable[float] | Callable[[str], float] | None, default=None
        Y-coordinates for each row of data. If None, uses the column names converted to float.
        If callable, applies the function to each column name to generate y values.
    cmap : Colormap | None, default=None
        Colormap to use for the plot. If None, uses viridis.
    norm : Normalize | None, default=None
        Normalization to apply to the colormap.
    N : int, default=256
        Number of color levels in the colormap.
    yscale : str, default="log"
        Scale for the y-axis, typically "log" or "linear".
    invert_y : bool, default=False
        Whether to invert the y-axis direction.
    edge_lines : bool | dict, default=True
        Whether to draw horizontal lines at the edges between chiclets.
        If a dict, passes those parameters to the axhline call.
    Returns
    -------
    QuadMesh
        The mesh object created by pcolormesh for further customization.

    """

    data_df = df.drop(x)
    if y_values is None:
        y_values = np.array([float(col) for col in data_df.columns])
    elif callable(y_values):
        y_values = np.array([y_values(col) for col in data_df.columns])
    elif isinstance(y_values, Iterable):
        y_values = np.array(y_values)

    x_vals = df[x]

    data = data_df.to_numpy().T

    diffs = np.diff(y_values)
    y_lower = y_values[0] - diffs[0] / 2
    y_upper = y_values[-1] + diffs[-1] / 2

    ymin = max(y_lower, y_values[0] / 2)
    ymax = min(y_upper, y_values[-1] * 2)

    mesh = ax.pcolormesh(
        x_vals, y_values, data, cmap=cmap or uplt.Colormap("viridis"), norm=norm, N=N
    )

    ax.format(yscale=yscale, ylim=(ymin, ymax), xlabel="Residue number")

    if edge_lines:
        kwargs = dict(color="k", lw=0.5, ls="-")
        if isinstance(edge_lines, dict):
            kwargs |= edge_lines
        edges = uplt.edges(y_values)
        for edge in edges:
            ax.axhline(edge, **kwargs)  # type: ignore

    if invert_y:
        ax.invert_yaxis()

    return mesh


def chiclet_categorical(
    ax: uplt.Axes,
    df: pl.DataFrame,
    x: str = "r_number",
    cmap: Colormap | None = None,
    norm: Normalize | None = None,
    N: int = 256,
    labels: list[str] | str | Callable[[str], str] | None = None,
    invert_y: bool = True,
    edge_lines: bool | dict = True,
) -> QuadMesh:
    """
    Create a categorical heatmap (chiclet plot) using pcolormesh.
    This function creates a heatmap where each row represents a categorical variable (ie protein states, mutants, etc.)
    and each column typically represents a position (like a residue number in a protein).
    Parameters
    ----------
    ax : uplt.Axes
        The axes on which to draw the plot.
    df : pl.DataFrame
        The dataframe containing the data to plot. Should include the column specified by `x`.
    x : str, default="r_number"
        The name of the column in `df` that contains the x-axis values.
    cmap : matplotlib.colors.Colormap or None, default=None
        The colormap to use. If None, defaults to viridis.
    norm : matplotlib.colors.Normalize or None, default=None
        The normalization to use for the colormap.
    N : int, default=256
        Number of color levels in the colormap.
    labels : list[str] or str or Callable[[str], str] or None, default=None
        The y-axis labels for the plot:
        - If None, uses column names from the dataframe
        - If str, formats each column name using the string
        - If callable, applies the function to each column name
        - If list, uses the provided list (must match column count)
    invert_y : bool, default=False
        If True, inverts the y-axis so the first category is at the top.
    edge_lines : bool or dict, default=True
        If True, draws horizontal lines between categories.
        If a dict, passes the dict as keyword arguments to ax.axhline().
    Returns
    -------
    QuadMesh
        The mesh object created by pcolormesh.
    Notes
    -----
    The dataframe columns (except the x column) become the categories (rows) in the plot.
    The x column values become the x-axis values (columns) in the plot.
    """

    data_df = df.drop(x)

    if labels is None:
        labels = data_df.columns
    elif isinstance(labels, str):
        labels = [labels.format(col) for col in data_df.columns]
    elif callable(labels):
        labels = [labels(col) for col in data_df.columns]
    elif isinstance(labels, list):
        if len(labels) != len(data_df.columns):
            raise ValueError(
                f"Length of labels ({len(labels)}) does not match number of columns in dataframe ({len(data_df.columns)})"
            )

    x_vals = df[x]
    y_vals = np.arange(len(labels)) + 1
    data = data_df.to_numpy().T
    mesh = ax.pcolormesh(
        x_vals, y_vals, data, cmap=cmap or uplt.Colormap("viridis"), norm=norm, N=N
    )

    ylocs = uplt.Locator("fixed", locs=y_vals)
    yfmt = uplt.Formatter(labels)
    ax.format(
        yscale="linear",
        xlabel="Residue number",
        ylocator=ylocs,
        yformatter=yfmt,
    )

    if edge_lines:
        kwargs = dict(color="k", lw=0.5, ls="-")
        if isinstance(edge_lines, dict):
            kwargs |= edge_lines
        edges = uplt.edges(y_vals)
        for edge in edges:
            ax.axhline(edge, **kwargs)  # type: ignore

    if invert_y:
        ax.invert_yaxis()

    return mesh


def mpl_peptides(
    ax: Axes,
    df: pl.DataFrame,
    qty: PlotQty,
    start: str = "start",
    end: str = "end",
    value: str = "value",
    colorbar: bool | dict = True,
    annotate: bool | dict = True,
    wrap: int | None = None,
) -> list[Patch]:
    N = len(df)

    colors = qty.get_colors(df[value])

    if wrap is None:
        wrap = find_wrap(df["start", "end"], margin=4, step=1, wrap_limit=200)

    y = np.arange(N) % wrap
    y = wrap - y
    kw = {"linewidth": 0.5, "linestyle": "-", "edgecolor": "k"}
    patches = []
    for idx, (y_pos, _start, _end, color) in enumerate(zip(y, df[start], df[end], colors)):
        width = _end - _start + 1
        rect = Rectangle((_start - 0.5, y_pos - 1), width, 1, facecolor=color, **kw)
        patch = ax.add_patch(rect)
        patches.append(patch)

        if annotate:
            # Annotate the rectangle with the index number
            center_x = _start + width / 2
            center_y = y_pos - 0.5
            annotate_defaults = dict(color="k", fontsize=10, ha="center", va="center")
            if isinstance(annotate, dict):
                annotate_defaults.update(annotate)
            text = str(idx)
            ax.annotate(text, (center_x, center_y), **annotate_defaults)

    # v_start, v_stop = zip(*intervals)
    vlower, vupper = df[start].min(), df[end].max() + 1
    width = vupper - vlower
    pad = 0.05
    xlim = (vlower - pad * width, vupper + 1 + pad * width)
    # pad x lim with range
    ax.set_xlim(*xlim)
    ax.set_ylim(0, wrap)
    ax.set_yticks([])
    ax.set_xlabel("Residue Number")

    if colorbar:
        cbar_kw = dict(width="4mm")
        if isinstance(colorbar, dict):
            cbar_kw.update(colorbar)

        if isinstance(ax, (uplt.Axes, uplt.SubplotGrid)):
            cbar = ax.colorbar(qty.mappable, label=qty.label, **cbar_kw)
        else:
            divider = make_axes_locatable(ax)
            size = uplt.utils.units(cbar_kw["width"])
            pad = cbar_kw.get("pad", 0.05)
            cax = divider.append_axes("right", size=size, pad=uplt.utils.units(pad))
            cmap = cm.get_cmap(qty.cmap, N)
            norm = Normalize(vmin=qty.norm.vmin, vmax=qty.norm.vmax)
            cbar = Colorbar(cax, cmap=cmap, norm=norm)
            cbar.set_label(qty.label)
            cax.grid(False)

    return patches


def scatter(
    ax: Axes,
    df: pl.DataFrame,
    qty: PlotQty,
    r_number: str = "r_number",
    value: str = "value",
    value_sd: str | None = None,
) -> Axes:
    ax.scatter(
        df[r_number],
        df[value] * qty.scale_factor,
        c=df[value] * qty.scale_factor,
        cmap=qty.cmap,
        norm=qty.norm,
    )
    ax.set_ylabel(qty.label)
    ax.set_xlabel("Residue number")

    value_sd = f"{value}_sd" if value_sd is None else value_sd
    if value_sd in df.columns:
        ax.errorbar(
            df[r_number],
            df[value] * qty.scale_factor,
            yerr=df[value_sd] * qty.scale_factor,
            fmt="o",
            ecolor="grey",
            color="none",
            zorder=-10,
        )

    return ax



from pathlib import Path
from string import Template
from instagibbs.__version__ import __version__
from hdxms_datasets import HDXDataSet

BASE_DIR = Path(__file__).parent
about_template = Template((BASE_DIR / "about.md").read_text())
subs = {"version": str(__version__)}
about_md = about_template.safe_substitute(subs)


# datasets which are available in memory
IN_MEMORY_DATASETS: dict[str, HDXDataSet] = {}



# %%
from __future__ import annotations
from dataclasses import dataclass
from typing import TYPE_CHECKING

import colorcet as cc
import matplotlib.cm as cm
import numpy as np
import tol_colors as tc
from matplotlib.colors import Colormap, Normalize, to_hex
from instagibbs.utils import rgb_to_hex

if TYPE_CHECKING:
    from altair import Scale

NO_COVERAGE = "#8c8c8c"


def make_cmap(cmap_source, bad_color: str = NO_COVERAGE) -> Colormap:
    """Create a colormap with bad values set."""
    cmap = cmap_source.copy()  # Make a copy to avoid modifying original
    cmap.set_bad(bad_color)
    return cmap


@dataclass(frozen=True)
class PlotQty:
    """
    A dataclass to hold the plotting quantity and its associated colormap and normalization.
    """

    name: str
    unit: str
    cmap: Colormap
    vmin: float = 0.0
    vmax: float = 1.0
    clip: bool = True
    scale_factor: float = 1.0

    @property
    def label(self) -> str:
        """
        Returns the label for the quantity, formatted as 'name (unit)'.
        """
        return f"{self.name} ({self.unit})"

    @property
    def mappable(self) -> cm.ScalarMappable:
        """
        Returns a ScalarMappable object for the colormap and normalization.
        """
        return cm.ScalarMappable(norm=self.norm, cmap=self.cmap)

    @property
    def color_bad(self) -> str:
        """
        Returns the color used for bad values (Null/NaN) in the colormap.
        """
        return to_hex(self.cmap.get_bad())  # type: ignore

    @property
    def norm(self) -> Normalize:
        return Normalize(vmin=self.vmin, vmax=self.vmax, clip=self.clip)

    @property
    def alt_scale(self) -> Scale:
        from altair import Scale

        N = 256
        domain = self.norm.vmin, self.norm.vmax
        hex_colors = rgb_to_hex(self.cmap(np.linspace(0, 1, N, endpoint=True), bytes=True))
        scale = Scale(domain=domain, range=hex_colors, clamp=True)  # type: ignore

        return scale

    def get_colors(self, values, bytes: bool = False):
        return self.cmap(self.norm(self.scale_factor * values), bytes=bytes)


RFU = PlotQty(
    name="RFU",
    unit="a.u.",
    cmap=make_cmap(cc.cm.gouldian),
    scale_factor=1.0,
)

DRFU = PlotQty(
    name="ΔRFU",
    unit="a.u.",
    cmap=make_cmap(cc.cm.diverging_bwr_20_95_c54),
    vmin=-0.5,
    vmax=0.5,
    scale_factor=1.0,
)

DG = PlotQty(
    name="ΔG",
    unit="kJ/mol",
    cmap=make_cmap(tc.tol_cmap("rainbow_PuRd")),  # type: ignore
    vmin=-40,
    vmax=-10,
    scale_factor=1e-3,  # convert J to kJ
)

DDG = PlotQty(
    name="ΔΔG",
    unit="kJ/mol",
    cmap=make_cmap(tc.tol_cmap("PRGn").reversed()),  # type: ignore
    vmin=-10,
    vmax=10,
    scale_factor=1e-3,  # convert J to kJ
)



"""
Load a set of HDX states and show as ViewCard
For running in (VSCode) interactive window

"""

from pathlib import Path

from hdxms_datasets import DataBase

from instagibbs.models import HDXState
from instagibbs.preprocess import load_hdx_dataset
from instagibbs.web.app import ViewCard, DataManager

# %%

root = Path(__file__).parent.parent
database_dir = root / "tests" / "test_data"
vault = DataBase(database_dir=database_dir)


DATASET = "HDX_0A55672B"
ds = vault.load_dataset(DATASET)

# %%
states = load_hdx_dataset(ds)
dm = DataManager(hdx_states=states)
ViewCard(dm)
# %%
